#!/usr/bin/env python
import ursgal
import os


class svm_1_0_0( ursgal.UNode ):
    """svm_1_0_0 UNode"""

    META_INFO = {
        'edit_version'       : 1.00,
        'name'               : 'Svm',
        'version'            : '1.0.0',
        'release_date'       : None,
        'engine_type' : {
            'validation_engine' : True,
        },
        'input_extensions'   : ['.csv'],
        'output_extensions'  : ['.csv'],
        'output_suffix'      : 'svm_validated',
        'create_own_folder'  : False,
        'in_development'     : True,  # do not show in UNode overview
        'include_in_git'     : True,
        'distributable'      : True,
        'utranslation_style' : 'svm_style_1',
        'engine' : {
            'platform_independent' : {
                'arc_independent' : {
                    'exe' : 'svm_1_0_0.py',
                },
            },
        },
        'citation' : \
            ''
    }

    def __init__(self, *args, **kwargs):
        super(svm_1_0_0, self).__init__(*args, **kwargs)

    def preflight(self):
        '''
        Building the list of parameters that will be passed to the
        svm_1_0_0 main function.

        These parameters are stored in self.params['command_list']

        Returns:
                None
        '''
        if not self.params['input_file'].lower().endswith('.csv'):
            raise ValueError(
                '\nSVM input file must be a unified CSV file, '
                'but you specified: ' + self.params['input_file']
            )

        in_path = os.path.join(
            self.params['input_dir_path'],
            self.params['input_file']
        )

        out_path = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )

        self.params['command_list'] = [
            '/usr/bin/env',
            'python3',
            self.exe,
            '--input_csv',
            in_path,
            '--output_csv',
            out_path,
            '--kernel',
            self.params['kernel'],
            '--fdr_cutoff',
            str(self.params['fdr_cutoff']),
            '-c',
            str(self.params['translations']['svm_c_param']),
            #  '--mb_ram',
            #  str(self.params['available_RAM_in_MB']),
        ]

        if self.params.get('columns_as_features', False):
            self.params['command_list'].append(
                '--columns_as_features'
            )
            self.params['command_list'].append(
                ' '.join(self.params['columns_as_features'])
            )
        return

    def postflight(self):
        return



#!/media/lukas/storage/uni/forschungsmodul_proteomics/pypercolator/pypercolator/bin/python

'''
Parses a target/decoy search result file and adds a column called "estimated_FDR".
The CSV has to contain: a column with a quality score for each PSM (e-value, error probability etc.)
and a column called "Is decoy" indicating whether a PSM is decoy or target.

usage:
./add_estimated_fdr_1_0_0.py search_results.csv output_filename.csv name_of_my_quality_score_column
'''

import csv
import os.path
import sys


def calc_FDR ( PSM_count, false_positives ):
    '''
    calculate false discovery rate according to FDR Method 2
    (Käll et al. 2007) as explained by Jones et al. 2009
    '''
    true_positives  = PSM_count - (2 * false_positives)
    if true_positives <= 0:  # prevent FDR above 1. Not sure if this is done?
        return 1.0
    FDR = false_positives / ( false_positives + true_positives)
    return FDR


def main( input_file=None, output_file=None, score_colname=None, bigger_scores_better=False ):
    
    assert input_file is not None, "add_estimated_FDR requires an input file!"
    assert score_colname is not None, "add_estimated_FDR needs to know the quality score column's name!"

    if output_file is None:
        base, ext   = os.path.splitext( input_file )
        output_file = base + "_FDR" + ext

    PSM_count   = 0
    decoy_count = 0
    
    with open( input_file, 'r' ) as in_csv:
        with open( output_file, 'w', newline='' ) as out_csv:
            
            reader     = csv.DictReader( in_csv )

            assert "estimated_FDR" not in reader.fieldnames, '''
  Input file {0} to add_estimated_fdr already has a column called "estimated_FDR"!
            '''.format( input_file )

            print('''Sorting {0} by column "{1}" ...'''.format(input_file, score_colname) )

            sorted_csv = sorted( reader, key=lambda d: float(d[ score_colname ]), reverse=bigger_scores_better )
            header     = reader.fieldnames + [ "estimated_FDR" ]

            writer     = csv.DictWriter( out_csv, header )
            writer.writeheader()

            print('''Adding "estimated_FDR" column to {0} ...'''.format(input_file) )

            # Traverse identifications from lowest score to highest:
            for row in sorted_csv:
                PSM_count += 1

                if row['Is decoy'].lower() == 'true':
                    decoy_count += 1

                # calculate and store the estimated FDR (FDRest) for each identification according to FDR Method 2.
                row["estimated_FDR"]  = calc_FDR ( PSM_count, decoy_count )
                writer.writerow( row )


if __name__ == "__main__":
    main(
        input_file    = sys.argv[1],
        output_file   = sys.argv[2],
        score_colname = sys.argv[2],
    )


#!/usr/bin/env python
import ursgal
import importlib
import os
import sys
import pickle


class add_estimated_fdr_1_0_0(ursgal.UNode):
    '''
    add_estimated_fdr_1_0_0 UNode

    Parses a target/decoy search result file and adds a column called
    "estimated_FDR".

    The CSV must contain:

        * a column with a quality score for each PSM (e-value, error probability etc.)
        * a column called "Is decoy" indicating whether a PSM is decoy or target.

    Note:
        This function can be used to independently compare the
        performance of different quality scores (where performance
        is the ability to distinguish target PSMs from decoy PSMs).

    '''
    META_INFO = {
        'edit_version'       : 1.01,
        'name'               : 'Add Estimated FDR',
        'version'            : '1.0.0',
        'release_date'       : '2008-1-1',
        'engine_type' : {
            'validation_engine' : True
        },
        'input_extensions'   : ['.csv'],
        'output_extensions'  : ['.csv'],
        'output_suffix'      : 'withFDR',
        'in_development'     : False,
        'include_in_git'     : True,
        'distributable'      : True,
        'utranslation_style' : 'add_estimated_fdr_style_1',
        'engine' : {
            'platform_independent' : {
                'arc_independent' : {
                    'exe' : 'add_estimated_fdr_1_0_0.py',
                },
            },
        },
        'citation' :
        'An implementation of the target/decoy FDR estimation method '
            'described in: Lukas Kall, John D. Storey, Michael J. MacCoss and '
            'William Stafford Noble (2008) Assigning significance to peptides '
            'identified by tandem mass spectrometry using decoy databases.' ,
    }

    def __init__(self, *args, **kwargs):
        super(add_estimated_fdr_1_0_0, self).__init__(*args, **kwargs)

    def _execute(self):
        '''
        '''
        print('[ -ENGINE- ] Executing conversion ..')
        # self.time_point(tag = 'execution')
        add_fdr_main = self.import_engine_as_python_function()
        if self.params['input_file'].lower().endswith('.csv') is False:
            raise ValueError(
                'add_estimated_fdr_1_0_0 input file must be a CSV file!'
            )

        # if the user specified a specific score to use for FDR estimation, 
        # use that:
        if self.params['validation_score_field'] is not None and self.params['bigger_scores_better'] is not None:
            score_field = self.params['validation_score_field']
            bigger_is_better = self.params['bigger_scores_better']
        # otherwise, estimate the FDR based on the score 
        # of the last search engine:
        else:
            last_search_engine = self.get_last_search_engine(
                history=self.stats['history']
            )
            assert last_search_engine, '''
  Could not detect which search engine produced the input file.
  "add_estimated_fdr" requires a quality criterion (score) to calculate estimated FDRs.
  Please make sure your input file has a proper ursgal u.json file, or define the
  column name of that score using this syntax:
  >>> uc.params['validation_score_field'] = 'my_score_column'
  >>> uc.params['bigger_scores_better'] = False  # or True!
        '''
            print('''
[ WARNING ] Estimating FDR based on the raw "{}" scores since
[ WARNING ] "validation_score_field" and/or "bigger_scores_better"
[ WARNING ] were not manually specified in the UController.params.
            '''.format(last_search_engine))
            score_field = self.UNODE_UPARAMS['validation_score_field'][
                'uvalue_style_translation'][last_search_engine]
            bigger_is_better = self.UNODE_UPARAMS['bigger_scores_better'][
                'uvalue_style_translation'][last_search_engine]

        output_file = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )
        input_file = os.path.join(
            self.params['input_dir_path'],
            self.params['input_file']
        )

        add_fdr_main(
            input_file=input_file,
            output_file=output_file,
            score_colname=score_field,
            bigger_scores_better=bigger_is_better,
        )

        # self.print_execution_time(tag='execution')
        return output_file



#!/usr/bin/env python
import ursgal
import importlib
import os
import sys
import pickle
import csv
import copy
import re
import pprint
from collections import defaultdict as ddict

class compomics_utilities_4_11_5( ursgal.UNode ):
    """
    compomics_utilities_4_11_5 UNode

    Reference:
        Barsnes, H., Vaudel, M., Colaert, N., Helsens, K., Sickmann, A.,
        Berven, F. S., and Martens, L. (2011) compomics-utilities: an 
        open-source Java library for computational proteomics. BMC\
        Bioinformatics 12, 70

    Warning:
        Implementation still in beta-stage!

    """
    META_INFO = {
        'edit_version'       : 1.00,
        'name'               : 'Compomic Utilities',
        'version'            : '4.11.5',
        'release_date'       : None,
        'engine_type' : {
            'misc_engine'     : True
        },
        'output_suffix'     : 'compomics',
        'input_extensions'  : ['.txt', ],
        'output_extensions' : ['.csv', ],
        'distributable' : False,
        'include_in_git'    : False,
        'in_development'    : False,
        'utranslation_style'    : 'compomics_utilities_style_1',
        'engine': {
            'platform_independent' : {
                'arc_independent' : {
                    'exe'            : 'utilities-4.11.5.jar',
                    'url'            : '',
                    'zip_md5'        : '',
                },
            },
        },
        'citation' : \
        'Barsnes, H., Vaudel, M., Colaert, N., Helsens, K., Sickmann, A.,'
        'Berven, F. S., and Martens, L. (2011) compomics-utilities: an '
        'open-source Java library for computational proteomics. BMC'
        'Bioinformatics 12, 70'
    }

    def __init__(self, *args, **kwargs):
        super(compomics_utilities_4_11_5, self).__init__(*args, **kwargs)

    def set_up_command_list(self):
        '''
        java
        -cp
        utilities-4.11.5.jar
        com.compomics.util.experiment.identification.protein_inference.executable.PeptideMapping
        -p /media/plan-f/Shared/databases/Saccharomyces_cerevisiae/uniprot_version_20161016_download_date_20161213/uniprot-proteome%3AUP000002311.fasta
        test.txt
        test_out.txt

        '''

        if self.params['translations']['compomics_utility_name'] == 'com.compomics.util.experiment.identification.protein_inference.executable.PeptideMapping':
            command_list = [
                'java',
                '-cp',
                self.exe,
                self.params['translations']['compomics_utility_name'],
                '-p',
                self.params['translations']['database'],
                self.tmp_peptide_file,
                self.params['translations']['output_file_incl_path'],
            ]
        return command_list

    def preflight( self ):
        '''


        '''
        input_file_incl_path = os.path.join(
            self.params['input_dir_path'],
            self.params['input_file']
        )
        self.params['translations']['output_file_incl_path'] = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )
        if self.params['translations']['compomics_utility_name'] == 'com.compomics.util.experiment.identification.protein_inference.executable.PeptideMapping':

            # read csv and write txt input file...
            self.tmp_peptide_file = os.path.join(
                self.params['output_dir_path'],
                'peptides.txt'
            )
            with open(self.tmp_peptide_file, 'w') as io:
                peptides = set()
                for line_dict in csv.DictReader(open(input_file_incl_path,'r')):
                    appended = False
                    for aa_to_replace, replace_dict in sorted(self.params['translations']['aa_exception_dict'].items(), reverse=True):
                        if aa_to_replace in line_dict['Sequence']:
                            if aa_to_replace in ['O']:
                                if 'unimod_name' in replace_dict.keys():
                                    for r_pos, aa in enumerate(line_dict['Sequence']):
                                        if aa == aa_to_replace:
                                            index_of_aa = r_pos + 1
                                            unimod_name = replace_dict['unimod_name']
                                            new_mod = '{0}:{1}'.format(
                                                unimod_name,
                                                index_of_aa
                                            )
                                            if line_dict['Modifications'] == '':
                                                line_dict['Modifications'] += new_mod
                                            else:
                                                line_dict['Modifications'] += ';{0}'.format(
                                                    new_mod
                                                )
                                if len(replace_dict['original_aa']) == 1:
                                    line_dict['Sequence'] = line_dict['Sequence'].replace(
                                        aa_to_replace,
                                        replace_dict['original_aa'][0]
                                    )
                            elif aa_to_replace in ['J']:
                                repeat_count = line_dict['Sequence'].count(aa_to_replace)
                                #we need to do all combos for the positions of I and L in the seq...
                                positions = []
                                for index_of_aa, aa in enumerate(line_dict['Sequence']):
                                    if aa == aa_to_replace:
                                        positions.append( index_of_aa )
                                product_combos = []
                                new_peptides = set()
                                for combo_tuple in itertools.product(replace_dict['original_aa'],repeat=repeat_count):
                                    # product_combos.append(n)
                                    for pos, new_aa in enumerate(combo_tuple):
                                        index_of_aa = positions[pos]
                                        # text[:1] + 'Z' + text[2:]
                                        remapped_peptide = '{0}{1}{2}'.format(
                                            line_dict['Sequence'][:index_of_aa],
                                            new_aa,
                                            line_dict['Sequence'][index_of_aa+1:],
                                        )
                                        tmp_peptide_set.add(
                                            remapped_peptide
                                        )
                                        line_dict['Sequence'] = remapped_peptide
                                        csv_file_buffer.append( deepcopy(line_dict) )
                                        appended = True
                            else:
                                print(
                                    '''
                                    [ WARNING ] New not covered case of aa exception for: "{0}"
                                    [ WARNING ] Please adjust upeptide_mapper accordingly
                                    '''.format(aa_to_replace)
                                )
                                sys.exit(1)
                    if appended is False:
                        peptides.add(line_dict['Sequence'])
                for peptide in peptides:
                    print(peptide,file=io)       

        self.params['command_list'] = self.set_up_command_list()
        return self.params['translations']['output_file_incl_path']

    def postflight( self ):
        '''
        add pre and post amino acids to results and extend txt results..
        '''
        # read database and get full name and also the pre and post aa
        short_2_long_name_mapping = {}
        fasta_id_dict = {}
        for fastaID, sequence in ursgal.ucore.parse_fasta( open( self.params['translations']['database'], 'r' ) ):
            fasta_id_dict[fastaID] = sequence
            try:
                short_2_long_name_mapping[ fastaID.split('|')[1] ] = fastaID # uniprot style
            except:
                pass
        all_full_protein_names = list(fasta_id_dict.keys())
        buffered_maps = ddict(list)
        with open(self.params['translations']['output_file_incl_path'],'r') as io:
            for line in io.readlines():
                peptide, protein, start_pos = line.split(',')
                if protein in fasta_id_dict.keys():
                    full_protein_name = protein
                else:
                    #this is slow and inefficient, but we need the full name
                    # use pyahocorasick automaton?? :)))
                    if protein in short_2_long_name_mapping.keys():
                        full_protein_name = short_2_long_name_mapping[protein]
                    else:
                        for full_protein_name in all_full_protein_names:
                            if protein in full_protein_name:
                                short_2_long_name_mapping[protein] = full_protein_name
                                break

                sequence = fasta_id_dict[full_protein_name]
                start_pos = int(start_pos)
                stop_pos = start_pos + len(peptide)
                pre_pos = start_pos - 1
                if pre_pos < 0:
                    pre = '-'
                else:
                    pre = sequence[ pre_pos ]
                try:
                    post = sequence[ stop_pos ]
                except:
                    post = '-'

                buffered_maps[peptide].append(
                    {
                        'start' : start_pos + 1,
                        'end'   : stop_pos,
                        'pre'   : pre,
                        'post'  : post,
                        'id'    : full_protein_name
                    }
                )
        with open(self.params['translations']['output_file_incl_path'],'w') as io:
            for peptide, info_dict_list in buffered_maps.items():
                for info_dict in info_dict_list:
                    print(
                        '{0},{id},{start},{end},{pre},{post}'.format( 
                            peptide,
                            **info_dict
                        ),
                        file =io
                    )
        os.remove(self.tmp_peptide_file)
        return



#!/usr/bin/env python
import gzip
import xml.etree.ElementTree as ElementTree
import csv
import os
import sys


def main(mzid_file, csv_file):
    """Convert msgfplus mzid file to csv.

    Args:
        mzid_file (str): file path to input mzid
        csv_file (str): file path to output csv

    Raises:
        Exception: Description
    """
    if mzid_file.endswith('.gz'):
        open_func = gzip.open
    else:
        open_func = open

    headers = [
        'Raw data location',
        'Spectrum ID',
        'Spectrum Title',
        'Exp m/z',
        'Charge',
        'Sequence',
        'proteinacc_start_stop_pre_post_;',
        'MS-GF:DeNovoScore',
        'MS-GF:RawScore',
        'MS-GF:SpecEValue',
        'MS-GF:EValue',
        'Modifications',
        'Retention Time (s)',
        'Start',
        'Stop',
        'Calc m/z',
        'Is decoy',
        'rank'
    ]
    pep_lookup = {}
    with open_func(mzid_file) as fin, open(csv_file, 'wt') as fout:
        csv_writer = csv.DictWriter(fout, fieldnames=headers)
        csv_writer.writeheader()
        tree = ElementTree.iterparse(fin, events=('end', 'start'))
        for event, element in tree:
            if element.tag.endswith('}Inputs') and event == 'end':
                for e in element:
                    if e.attrib.get('id') == 'SID_1':
                        input_filepath = e.attrib['location']
                        input_filename = e.attrib['name']
            if element.tag.endswith('}Peptide') and event == 'end':
                key = element.attrib['id']
                if key not in pep_lookup:
                    pep_lookup[key] = {}
                    pep_lookup[key]['mods'] = []
                else:
                    raise Exception('Duplicate key {key}'.format(key=key))

                for e in element:
                    if e.tag.endswith('}PeptideSequence'):
                        seq = e.text
                        pep_lookup[key]['sequence'] = seq
                    elif e.tag.endswith('}Modification'):
                        pos = e.attrib['location']
                        delta = e.attrib['monoisotopicMassDelta']
                        for mod_e in e:
                            new = {}
                            new['pos'] = pos
                            new['delta'] = delta
                            new['name'] = mod_e.attrib['name']
                            pep_lookup[key]['mods'].append(new)

            if element.tag.endswith('}SpectrumIdentificationResult') and event == 'end':
                for ele in element[::-1]:
                    if ele.attrib.get('name', None) == 'spectrum title':
                        spec_title = ele.attrib['value']
                        spec_id = spec_title.split('.')[1]
                    elif ele.tag.endswith('}SpectrumIdentificationItem'):
                        attrib = ele.attrib
                        ref = attrib['peptide_ref']

                        peptide_data = pep_lookup[ref]

                        mods = []
                        for m in peptide_data['mods']:
                            pos  = m['pos']
                            delta = m['delta']
                            string = '{delta}:{pos}'.format(
                                delta=delta,
                                pos=pos
                            )
                            mods.append(string)

                        mods = ';'.join(mods)
                        input_filename_stem, ext = os.path.splitext(
                            input_filename
                        )
                        charge = attrib['chargeState']
                        new_psm = {
                            'Spectrum ID'      : spec_id,
                            'Spectrum Title'   : spec_title,
                            'Exp m/z'          : attrib['experimentalMassToCharge'],
                            'Charge'           : charge,
                            'Raw data location': input_filepath,
                            'Sequence'         : peptide_data['sequence'],
                            'rank'             : attrib['rank'],
                            'Modifications'    : mods,
                        }

                        for e in ele:
                            name = e.attrib.get('name', None)
                            if name in headers:
                                new_psm[name] = e.attrib['value']
                        csv_writer.writerow(new_psm)

if __name__ == '__main__':
    if len(sys.argv) == 3:
        main(sys.argv[1], sys.argv[2])



#!/usr/bin/env python3
# encoding: utf-8
'''

Test the csv2ssl engine for unified file

'''
import ursgal
import csv
import os


R = ursgal.UController()

csv2ssl_main = R.unodes['csv2ssl_1_0_0']['class'].import_engine_as_python_function()
input_csv = os.path.join(
    'tests',
    'data',
    'csv2ssl_1_0_0',
    'BSA1_csv2ssl_test_unified.csv'
)
output_csv = os.path.join(
    'tests',
    'data',
    'csv2ssl_1_0_0',
    'BSA1_csv2ssl_test_unified_converted.csv'
)
csv2ssl_main(
    input_file     = input_csv,
    output_file    = output_csv,
    score_column_name = 'MS-GF:EValue',
    score_type        = 'TANDEM EXPECTATION VALUE',
)

ident_list = [ ]
csv_in = csv.DictReader(open(input_csv, 'r'))
fieldnames = csv_in.fieldnames
for line_dict in csv_in:
    ident_list.append( line_dict )

out_list = [ ]
csv_out = csv.DictReader(open(output_csv, 'r'), delimiter='\t')
for line_dict in csv_out:
    out_list.append( line_dict )

def csv2ssl_test():
    for test_id, test_dict in enumerate(ident_list):
        yield csv2ssl, test_dict, test_id


def csv2ssl( test_dict, test_id ):
    assert os.path.isfile(output_csv)

    for key in fieldnames:
        if key.startswith('Expected'):
            expected_value = test_dict[key]
            test_value = out_list[test_id][key[9:]]
            print(test_value, expected_value)
            assert test_value == expected_value


if __name__ == '__main__':
    print(__doc__)
    for test_id, test_dict in enumerate(ident_list):
        csv2ssl(test_dict, test_id)



#!/usr/bin/env python
'''
Sanitize the result csvs

usage:
    ./sanitize_csv_1_0_0.py <input_file> <output_file> <validation_score_field> <bigger_scores_better> <score_diff_threshold> <log10_threshold> <accept_conflicting_psms> <num_compared_psms> <remove_redundant_psms>


'''
import sys
import os
import csv
import ursgal
import math
import operator
from collections import defaultdict as ddict

if sys.platform != 'win32':
    csv.field_size_limit(sys.maxsize)


def main( 
    input_file=None, 
    output_file=None,
    grouped_psms=None,
    validation_score_field=None, 
    bigger_scores_better=None,
    score_diff_threshold=2.0,
    log10_threshold=True,
    accept_conflicting_psms=False,
    num_compared_psms=2,
    remove_redundant_psms=False,
):
    '''
    Spectra with multiple PSMs are sanitized, i.e. only the PSM with best PEP score is accepted
    and only if the best hit has a PEP that is at least two orders of magnitude smaller than the others
    '''

    # un = ursgal.UNode()
    # grouped_psms = un._group_psms( input_file, validation_score_field=validation_score_field, bigger_scores_better=bigger_scores_better )

    if grouped_psms is None:
        grouped_psms = group_psms( input_file, validation_score_field=validation_score_field, bigger_scores_better=bigger_scores_better )

    all_line_dicts = []
    for spec_title, grouped_psm_list in grouped_psms.items():
        spec_line_dicts = []
        best_score = None
        psm_names = set()
        for n, grouped_psm in enumerate(grouped_psm_list):
            score, line_dict = grouped_psm
            if n == 0:
                best_score = score
                psm_names.add(line_dict['Sequence']+line_dict['Modifications']+line_dict['Charge'])
                spec_line_dicts.append(line_dict)
            elif n < num_compared_psms:
                psm = line_dict['Sequence']+line_dict['Modifications']+line_dict['Charge']
                if psm in psm_names and remove_redundant_psms is True:
                    continue
                if log10_threshold is True:
                    if abs(math.log10(best_score)) - abs(math.log10(score)) >= score_diff_threshold:
                        break
                    else:
                        spec_line_dicts.append(line_dict)
                        psm_names.add(psm)
                else:
                    if abs(best_score) - abs(score) >= score_diff_threshold:
                        break
                    else:
                        spec_line_dicts.append(line_dict)
                        psm_names.add(psm)
        if accept_conflicting_psms is False and len(spec_line_dicts) >= 2:
            continue   
        else:
            all_line_dicts.extend(spec_line_dicts)


    csv_kwargs = {}
    if sys.platform == 'win32':
        csv_kwargs['lineterminator'] = '\n'
    else:
        csv_kwargs['lineterminator'] = '\r\n'

    output_file_object = open(output_file,'w')

    with open( input_file, 'r' ) as in_file:
        csv_input  = csv.DictReader( in_file )
        fieldnames = csv_input.fieldnames
    in_file.close()
        
    csv_output = csv.DictWriter(
        output_file_object,
        list(fieldnames),
        **csv_kwargs
    )
    csv_output.writeheader()
    for line_dict in all_line_dicts:
        csv_output.writerow( line_dict )
    output_file_object.close()
    return output_file

def group_psms(input_file, validation_score_field=None, bigger_scores_better=None):
    '''
    reads an input csv and returns a defaultdict with the spectrum title
    mapping to a sorted list of tuples containing each
    a) score (from validation_score_field) and
    b) the whole line dict
    '''
    print('[ GROUPING ] Parsing {0}'.format( input_file ))

    if validation_score_field is None:
        search_engine = self.get_last_search_engine( history = self.stats['history'] )
        assert search_engine, 'Can\'t convert results from no specified search engine.'
        assert 'multiple engines:' not in search_engine, 'Can\'t convert merged results from multiple different engines.'
        validation_score_field = self.UNODE_UPARAMS['validation_score_field']['uvalue_style_translation'][search_engine]

    if bigger_scores_better is None:
        bigger_scores_better = self.UNODE_UPARAMS['bigger_scores_better']['uvalue_style_translation'][search_engine]

    grouped_psms = ddict(list)
    opened_file = open( input_file, 'r')
    csv_dict_reader_object = csv.DictReader(
        row for row in opened_file if not row.startswith('#')
    )
    n = 0

    for n, line_dict in enumerate(csv_dict_reader_object):
        assert validation_score_field in line_dict.keys(), \
            '''defined validation_score_field {0} is not found,
            please check/add it to uparams.py['validation_score_field']'''.format(validation_score_field)


        grouped_psms[ line_dict[ 'Spectrum Title' ] ].append(
            (
                float(
                    line_dict[ validation_score_field ]
                ),
                line_dict
            )
        )
    for spectrum_title in grouped_psms.keys():
        grouped_psms[ spectrum_title ].sort(
            key     = operator.itemgetter(0),
            reverse = bigger_scores_better
        )
    print(
        "[ GROUPING ] Grouped {0} PSMs into {1} unique spectrum titles".format(
            n,
            len( grouped_psms.keys())
        )
    )
    return grouped_psms


if __name__ == '__main__':
    if len(sys.argv) < 10:
        print(__doc__)
        sys.exit(1)

    main(
        input_file              = sys.argv[1],
        output_file             = sys.argv[2],
        validation_score_field  = sys.argv[3], 
        bigger_scores_better    = bool(sys.argv[4]),
        score_diff_threshold    = int(sys.argv[5]),
        log10_threshold         = bool(sys.argv[6]),
        accept_conflicting_psms = bool(sys.argv[7]),
        num_compared_psms       = int(sys.argv[8]),
        remove_redundant_psms   = bool(sys.argv[9]),
    )



#!/usr/bin/env python
import ursgal
import os
import os.path


class myrimatch_2_1_138( ursgal.UNode ):
    """
    Myrimatch UNode

    Myrimatch options:
    http://forge.fenchurch.mc.vanderbilt.edu/scm/viewvc.php/*checkout*/trunk/doc/index.html?root=myrimatch

    Reference:
    Tabb DL, Fernando CG, Chambers MC. (2007) MyriMatch: highly accurate tandem mass spectral peptide identification by multivariate hypergeometric analysis.
    """
    META_INFO = {
        'edit_version'                : 1.00,
        'name'                        : 'Myrimatch',
        'version'                     : '2.1.138',
        'release_date'                : None,
        'engine_type' : {
            'protein_database_search_engine' : True,
        },
        'input_extensions'            : ['.mzML'],
        'output_extensions'           : ['.mzid'],
        'in_development'              : False,
        'include_in_git'              : False,
        'distributable'               : True,
        'create_own_folder'           : True,
        'utranslation_style'          : 'myrimatch_style_1',
        'engine' : {
            'linux' : {
                '64bit' : {
                    'exe'            : 'myrimatch_2_1_138',
                    'url'            : '',
                    'zip_md5'        : 'bf5df092579fe3f3d1364c835fe2f3ea',
                    'additional_exe' : [],
                },
            },
        },
        'citation' : \
            'Tabb DL, Fernando CG, Chambers MC. (2007) MyriMatch: highly '\
            'accurate tandem mass spectral peptide identification by '\
            'multivariate hypergeometric analysis.',
    }

    def __init__(self, *args, **kwargs):
        super(myrimatch_2_1_138, self).__init__(*args, **kwargs)
        pass

    def preflight( self ):
        '''
        Formatting the command line
        '''
        input_file = os.path.join(
            self.params['input_dir_path'],
            self.params['input_file']
        )
        if input_file.lower().endswith('.mzml') or \
            input_file.lower().endswith('.mzml.gz'):
            self.params['translations']['mzml_input_file'] = input_file
        elif input_file.lower().endswith('.mgf'):
            # use dark magic to find the mzml file that corresponds to the mgf:
            self.params['translations']['mzml_input_file'] = \
                self.meta_unodes['ucontroller'].get_mzml_that_corresponds_to_mgf( input_file )
            self.print_info(
                ('Since MyriMatch cannot read MGF input files, the '
                 'corresponding mzML file {0} will be used instead.'
                ).format( os.path.abspath(self.params['translations']['mzml_input_file']) ),
                caller="INFO" )
        else:
            raise Exception('MyriMatch input spec file must be in mzML format!')

        self.params['translations']['output_file_incl_path'] = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )

        self.params['translations']['myrimatch_ions_to_search'] = []
        for ion in ['a', 'b', 'c', 'x', 'y', 'z']:
            if ion in self.params['translations']['score_ion_list']:
                self.params['translations']['myrimatch_ions_to_search'].append(ion)
        self.params['translations']['myrimatch_ions_to_search'] = 'manual:'+','.join(self.params['translations']['myrimatch_ions_to_search'])

        self.params['translations']['myrimatch_config_file_path'] = \
            '{output_file_incl_path}_myrimatch_params.cfg'.format( **self.params['translations'])
        self.created_tmp_files.append( self.params['translations']['myrimatch_config_file_path'])
        self.write_param_file()

        self.params['command_list'] = [
            self.exe,
            '-cfg', '{myrimatch_config_file_path}'.format(**self.params['translations']),
            '{mzml_input_file}'.format(**self.params['translations']),
            '-workdir', '{output_dir_path}'.format(**self.params),
            '-cpus', '{cpus}'.format(**self.params['translations']),
            # '-dump'
        ]


    def write_param_file( self ):
        ''' Writes a file containing all parameters for the search '''

        config_file= open(self.params['translations']['myrimatch_config_file_path'], 'w') #param file must be in execution folder
        self.params['translations']['myrimatch_static_mods'] = ''
        self.params['translations']['myrimatch_dynamic_mods'] = ''

        if self.params['translations']['label'] == '15N':
            for aminoacid, N15_Diff in ursgal.ukb.DICT_15N_DIFF.items():
                existing = False
                for mod in self.params[ 'mods' ][ 'fix' ]:
                    if aminoacid == mod[ 'aa' ]:
                        mod[ 'mass' ] += N15_Diff
                        existing = True
                if existing == True:
                    continue
                self.params['translations']['myrimatch_static_mods']+= '{0} {1} '.format( aminoacid, N15_Diff )

        for mod in self.params[ 'mods' ][ 'fix' ]:
            self.params['translations']['myrimatch_static_mods'] += '{0} {1} '.format( mod[ 'aa' ], mod[ 'mass' ] )

        characters = ['$', '%', '^', '&', '*', '_', '<', '>', '|', ':', ';' ]
        for n, mod in enumerate( self.params[ 'mods' ][ 'opt' ] ):
            if n >= len(characters)-1:
                print(
                    '''
                    [ WARNING ] To many potential modifications. Ran out of characters.
                    [ WARNING ] Continues without modification: {0}'''.format(mod)
                )
                continue
            if mod[ 'pos' ] != 'any':
                print(
                    '''
                    [ WARNING ] myrimatch does only support "any" as position of modifications.
                    [ WARNING ] Continues without modification: {0}'''.format(mod)
                )
                continue
            self.params['translations']['myrimatch_dynamic_mods'] += '{0} {1} {2} '.format( mod[ 'aa' ], characters[ n ], mod[ 'mass' ] )

        print(
            '''
            [ WARNING ] precursor_mass_tolerance_plus and precursor_mass_tolerance_minus
            [ WARNING ] need to be combined for Myrimatch (use of symmetric tolerance window).
            [ WARNING ] The arithmetic mean is used.
            '''
        )
        self.params['translations']['precursor_mass_tolerance'] = ( float(self.params['translations']['precursor_mass_tolerance_plus']) + \
                                                    float(self.params['translations']['precursor_mass_tolerance_minus']) ) \
                                                / 2.0

        # temporary fix: # this should not be required anymore :)
        self.params['translations']['database'] = os.path.abspath( self.params['translations']['database'] )

        print('''
ClassSizeMultiplier = "{myrimatch_class_size_multiplier}"
CleavageRules = "{enzyme}"
ComputeXCorr = "{compute_xcorr}"
DecoyPrefix = "{decoy_tag}"
DynamicMods = "{myrimatch_dynamic_mods}"
EstimateSearchTimeOnly = "0"
FragmentMzTolerance = "{frag_mass_tolerance}{frag_mass_tolerance_unit}"
FragmentationAutoRule = "false"
FragmentationRule = "{myrimatch_ions_to_search}"
KeepUnadjustedPrecursorMz = "0"
MaxDynamicMods = "{max_num_mods}"
MaxMissedCleavages = "{max_missed_cleavages}"
MaxPeakCount = "{max_accounted_observed_peaks}"
MaxPeptideLength = "{max_pep_length}"
MaxPeptideMass = "{precursor_max_mass} Da"
MaxPeptideVariants = "{max_pep_var}"
MaxResultRank = "{num_match_spec}"
MinMatchedFragments = "{min_required_matched_peaks}"
MinPeptideLength = "{min_pep_length}"
MinPeptideMass = "{precursor_min_mass}"
MinResultScore = "{min_output_score}"
MinTerminiCleavages = "{semi_enzyme}"
MonoPrecursorMzTolerance = "{precursor_mass_tolerance}{precursor_mass_tolerance_unit}"
MonoisotopeAdjustmentSet = "{precursor_isotope_range}"
NumBatches = "{batch_size}"
NumChargeStates = "{precursor_max_charge}"
NumIntensityClasses = "{myrimatch_num_int_classes}"
NumMzFidelityClasses = "{myrimatch_num_mz_fidelity_classes}"
OutputSuffix = ""
OutputFormat = "mzIdentML"
PrecursorMzToleranceRule = "{precursor_mass_type}"
ProteinDatabase = "{database}"
ProteinSamplingTime = "{myrimatch_prot_sampl_time}"
SpectrumListFilters = "peakPicking true 2-"
StaticMods = "{myrimatch_static_mods}"
StatusUpdateFrequency = "5"
TicCutoffPercentage = "{myrimatch_tic_cutoff}"
UseSmartPlusThreeModel = "{myrimatch_smart_plus_three}"
    '''.format(**self.params['translations']), file=config_file)
        config_file.close()
        return

    def postflight(self):
        '''
        renaming MyriMatch's output file to our desired output file name
        '''
        if self.params['prefix'] is None:
            out_file_root = os.path.join(
                self.params['output_dir_path'],
                self.params['file_root']
            )
        else:
            out_file_root = os.path.join(
                self.params['output_dir_path'],
                self.params['file_root'].replace(
                    self.params['prefix'] + '_',
                    ''
                )
            )

        # when the input file is gzipped,
        # myrimatch does not trim the "mzml"-part of
        # the file extension, so sometimes the output
        # file might end with .mzML.mzid ...
        # in any case, we rename it to our desired name
        possible_outfiles = [
            out_file_root + '.mzid',  # most likely case, unless input was gzipped
            out_file_root + '.mzML.mzid',
            out_file_root + '.mzml.mzid',
            out_file_root + '.MZML.mzid',
        ]
        desired_output_name = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )

        for possible_outfile in possible_outfiles:
            if os.path.exists(possible_outfile):
                os.rename(
                    possible_outfile,
                    desired_output_name
                )
                break

        return



#!/usr/bin/env python3
# encoding: utf-8
'''

Test the ucore.digest function

Rules:
'enzyme' : {
    'argc' :            ('R', 'C'),
    'aspn' :            ('D', 'N'),
    'chymotrypsin' :    ('FMWY', 'C'),
    'chymotrypsin_p' :  ('FMWY', 'C'),
    'clostripain' :     ('R', 'C'),
    'cnbr' :            ('M', 'C'),
    'elastase' :        ('AGILV', 'C'),
    'formicacid' :      ('D', 'C'),
    'gluc' :            ('DE', 'C'),
    'gluc_bicarb' :     ('E', 'C'),
    'iodosobenzoate' :  ('W', 'C'),
    'lysc' :            ('K', 'C'),
    'lysc_p' :          ('K', 'C'),
    'lysn' :            ('K', 'N'),
    'lysn_promisc' :    ('AKRS', 'N'),
    'pepsina' :         ('FL', 'C'),
    'protein_endopeptidase' : ('P', 'C'),
    'staph_protease' :  ('E', 'C'),
    'trypsin' :         ('KR', 'C'),
    'trypsin_p' :       ('KR', 'C'),
    'trypsin/cnbr' :    ('KRM', 'C'),
    'trypsin/gluc' :    ('DEKR', 'C'),
}


'''
import ursgal.ucore
import os


TESTS = [
    {
        'sequence' : 'RKRK',
        'result'   : ['R','K','R','K'],
        'enzyme'   : ('KR', 'C'),
    },
    {
        'sequence' : 'IKI',
        'result'   : ['IK','I',],
        'enzyme'   : ('KR', 'C'),
    },
    {
        'sequence' : 'FMWY',
        'result'   : ['F','M','W','Y'],
        'enzyme'   : ('FMWY', 'C'),
    },
    {
        'sequence' : 'RKRK',
        'result'   : ['RK','RK'],
        'enzyme'   : ('K', 'C'),
    },
    {
        'sequence' : 'RKRK',
        'result'   : ['R','KR', 'K'],
        'enzyme'   : ('K', 'N'),
    },
    {
        'sequence' : 'PEPTID',
        'result'   : ['P','EP','TID'],
        'enzyme'   : ('P', 'C'),
    },
]

def digest_test():
    for test_id, test_dict in enumerate(TESTS):
        yield digest, test_dict

def digest( test_dict ):
    '''
    peptide_regex(self, database, protein_id, peptide):
    '''
    digest_result = ursgal.ucore.digest(
        test_dict['sequence'],
        test_dict['enzyme'],
        no_missed_cleavages = True
    )

    assert digest_result == test_dict['result']



if __name__ == '__main__':
    print(__doc__)
    for test_id, test_dict in enumerate(TESTS):
        digest(test_dict)



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
import os

R = ursgal.UController()

TESTS = [
    {
        'params' : {'database' : os.path.join(os.sep,'tmp','..','mo_fasta.fasta')} ,
        'output' : {'database' : os.path.join(os.path.abspath( os.sep ),'mo_fasta.fasta')}
    },
    {
        'params' : {'mosh' : os.path.join(os.sep,'tmp','..','mo_fasta.fasta')},
        'input_kwargs' : {'param_keys' : ['mosh']},
        'output' : {'mosh' : os.path.join(os.path.abspath( os.sep ),'mo_fasta.fasta')}
    },
]


def abs_paths_for_specific_keys_test():
    for test_id, test_dict in enumerate(TESTS):
        yield abs_paths_for_specific_keys, test_dict


def abs_paths_for_specific_keys( test_dict ):
    if 'input_kwargs' not in test_dict.keys():
        test_dict[ 'input_kwargs' ] = {}
    out_put = R.abs_paths_for_specific_keys(
        test_dict['params'],
        **test_dict['input_kwargs']
    )
    assert out_put == test_dict['output'], 'Abspaths {0} failed'.format(
        test_dict
    )


if __name__ == '__main__':
    for test_id, test_dict in enumerate(TESTS):
        abs_paths_for_specific_keys( test_dict )



#!/usr/bin/env python
import ursgal
import importlib
import os
import sys
import pickle
import shutil

class csv2ssl_1_0_0( ursgal.UNode ):
    """csv2ssl_1_0_0 UNode"""

    META_INFO = {
        'edit_version'           : 1.00,
        'name'                   : 'Convert CSV to SSL',
        'version'                : '1.0.0',
        'release_date'           : None,
        'engine_type' : {
            'converter'     : True
        },
        'output_extensions' : ['.ssl'],
        'input_extensions'  : ['.csv'],
        'output_suffix'    : 'converted',
        'in_development'   : False,
        'utranslation_style'    : 'csv2ssl_style_1',
        'include_in_git'   : True,
        'distributable'      : True,
        'engine': {
            'platform_independent' : {
                'arc_independent' : {
                    'exe'     : 'csv2ssl_1_0_0.py',
                },
            },
        },
        'citation' : '',
    }


    def __init__(self, *args, **kwargs):
        super(csv2ssl_1_0_0, self).__init__(*args, **kwargs)

    def _execute( self ):
        '''
        Result files (.csv) are converted to spectrum sequence list (.ssl) files.
        These .ssl can be used as input files for BiblioSpec.

        Input file has to be a .csv

        Creates a _converted.csv file and returns its path.
        '''

        print('[ -ENGINE- ] Executing conversion ..')
        self.time_point(tag = 'execution')
        csv2ssl_main = self.import_engine_as_python_function()
        if self.params['input_file'].lower().endswith('.csv') is False:
            raise ValueError('Trying to convert a non-csv file')

        output_file = os.path.join(
                self.params['output_dir_path'],
                self.params['output_file']
            )
        input_file  = os.path.join(
                self.params['input_dir_path'],
                self.params['input_file']
            )

        csv2ssl_main(
            input_file     = input_file,
            output_file    = output_file,
            score_column_name    = self.params['translations']['ssl_score_column_name'],
            score_type           = self.params['translations']['ssl_score_type'],
        )

        self.print_execution_time(tag='execution')
        return output_file



#!/usr/bin/env python3
# encoding: utf-8


import ursgal
import glob
import os.path
import sys
import time


def main(fasta_database, class_version):
    '''

    Example script to demonstrate speed and memory efficiency of the new 
    upeptide_mapper.

    Specify fasta_database and class_version as input.

    usage:
        ./complete_proteome_match.py <fasta_database> <class_version>

    Class versions
        * UPeptideMapper_v2
        * UPeptideMapper_v3
        * UPeptideMapper_v4

    '''

    input_params = {
        'database': sys.argv[1],
    }

    uc = ursgal.UController(
        params=input_params
    )

    print('Parsing fasta and digesting sequences')
    peptides = set()
    digest_start = time.time()
    for fastaID, sequence in ursgal.ucore.parse_fasta(open(input_params['database'], 'r')):
        tryptic_peptides = ursgal.ucore.digest(
            sequence,
            ('KR', 'C'),
            # no_missed_cleavages = True
        )
        for p in tryptic_peptides:
            if 6 <= len(p) <= 40:
                peptides.add(p)
    print(
        'Parsing fasta and digesting sequences took {0:1.2f} seconds'.format(
            time.time() - digest_start
        )
    )

    if sys.platform == 'win32':
        print(
            '[ WARNING ] pyahocorasick can not be installed via pip on Windwows at the moment\n'
            '[ WARNING ] Falling back to UpeptideMapper_v2'
        )
        class_version = 'UPeptideMapper_v2'

    upapa_class = uc.unodes['upeptide_mapper_1_0_0']['class'].import_engine_as_python_function(
        class_version
    )

    print('Buffering fasta and mapping {0} peptides'.format(len(peptides)))
    map_start = time.time()

    if class_version == 'UPeptideMapper_v2':
        peptide_mapper = upapa_class(word_len=6)
        fasta_lookup_name = peptide_mapper.build_lookup_from_file(
            input_params['database'],
            force=False,
        )
        args = [
            list(peptides),
            fasta_lookup_name
        ]
    elif class_version == 'UPeptideMapper_v3':
        peptide_mapper = upapa_class(input_params['database'])
        fasta_lookup_name = peptide_mapper.fasta_name
        args = [
            list(peptides),
            fasta_lookup_name
        ]
    elif class_version == 'UPeptideMapper_v4':
        peptide_mapper = upapa_class(input_params['database'])
        args = [
            list(peptides)
        ]
    p2p_mappings = peptide_mapper.map_peptides(*args)
    print(
        'Buffering fasta and mapping {0} peptides took {1:1.2f} seconds'.format(
            len(peptides),
            time.time() - map_start
        )
    )
    if len(p2p_mappings.keys()) == len(peptides):
        print('All peptides have been mapped!')
    else:
        print('WARNING: Not all peptide have been mapped')

if __name__ == "__main__":
    if len(sys.argv) < 3:
        print(main.__doc__)
        sys.exit(1)
    main(sys.argv[1], sys.argv[2])



#!/usr/bin/env python3
'''
'''
import os
import ursgal


class pyqms_1_0_0(ursgal.UNode):
    """
    pyQms- molecule qunatification node.
    """
    META_INFO = {
        'edit_version': 1.00,
        'name': 'pyqms',
        'version': '1.0.0',
        'release_date': None,
        'engine_type': {
            'quantification_engine': True
        },
        'utranslation_style': 'pyqms_style_1',
        'citation': 'Leufken J, Niehues A, Sarin LP, Wessel F, Hippler M, Leidel SA, Fufezan C (2017) pyQms enables universal and accurate quantification of mass spectrometry data',
        'input_extensions': ['.mzML'],
        'output_extensions': ['.csv'],
        'create_own_folder': True,
        'in_development': False,
        'include_in_git': False,
        'distributable': False,
        # Download Information
        'engine': {
            'platform_independent': {
                'arc_independent': {
                    'exe': 'pyqms_1_0_0.py',
                },
            },
        },
    }

    def __init__(self, *args, **kwargs):
        """."""
        super(pyqms_1_0_0, self).__init__(*args, **kwargs)

    def _execute(self):
        """run."""
        print('[ -ENGINE- ] Executing quantification ..')
        self.time_point(tag='execution')

        main = self.import_engine_as_python_function()

        output_file = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )

        pickle_file = os.path.join(
            self.params['output_dir_path'],
            os.path.splitext(self.params['output_file'])[0] + '_quant.pkl',
        )
        # multiple input mzML files
        if self.params['input_file'].endswith('.json'):
            mzml_files = []
            for fdict in self.params['input_file_dicts']:
                mzml_files.append(fdict['full'])
        else:  # single mzML file
            mzml_files = os.path.join(
                self.params['input_dir_path'],
                self.params['input_file']
            )
        if self.params['translations']['ms_level'] == 1:
            print(
                '''
                [ WARNING ] precursor_mass_tolerance_plus and precursor_mass_tolerance_minus
                [ WARNING ] need to be combined for pyQms (use of symmetric tolerance window).
                [ WARNING ] The arithmetic mean is used.
                '''
            )
            self.params['translations']['REL_MZ_RANGE'] = (
                float(self.params['translations']['precursor_mass_tolerance_plus']) +
                float(self.params['translations'][
                      'precursor_mass_tolerance_minus'])) / 2.0
            if self.params['translations']['precursor_mass_tolerance_unit'] == 'da':
                self.params['translations']['REL_MZ_RANGE'] = \
                    ursgal.ucore.convert_dalton_to_ppm(
                        self.params['translations']['REL_MZ_RANGE'],
                        base_mz=self.params['translations']['base_mz']
                )
        else:
            self.params['translations']['REL_MZ_RANGE'] = \
                self.params['translations']['frag_mass_tolerance']
            if self.params['translations']['frag_mass_tolerance_unit'] == 'da':
                self.params['translations']['REL_MZ_RANGE'] = \
                    ursgal.ucore.convert_dalton_to_ppm(
                        self.params['translations']['REL_MZ_RANGE'],
                        base_mz=self.params['translations']['base_mz']
                )
        self.params['translations']['REL_MZ_RANGE'] = self.params[
            'translations']['REL_MZ_RANGE'] * 1e-6

        pyqms_params = {
            'PERCENTILE_FORMAT_STRING':
                self.params['translations']['label_percentile_format_string'],
            'M_SCORE_THRESHOLD':
                self.params['translations']['m_score_cutoff'],
            'ELEMENT_MIN_ABUNDANCE':
                self.params['translations']['min_element_abundance'],
            'MIN_REL_PEAK_INTENSITY_FOR_MATCHING':
                self.params['translations'][
                    'min_rel_peak_intensity_for_matching'],
            'REQUIRED_PERCENTILE_PEAK_OVERLAP':
                self.params['translations'][
                    'required_percentile_peak_overlap'],
            'MINIMUM_NUMBER_OF_MATCHED_ISOTOPOLOGUES':
                self.params['translations'][
                    'min_number_of_matched_isotopologues'],
            'INTENSITY_TRANSFORMATION_FACTOR':
                self.params['translations']['intensity_transformation_factor'],
            'UPPER_MZ_LIMIT':
                self.params['translations']['upper_mz_limit'],
            'LOWER_MZ_LIMIT':
                self.params['translations']['lower_mz_limit'],
            'MZ_TRANSFORMATION_FACTOR':
                self.params['translations']['mz_transformation_factor'],
            'REL_MZ_RANGE':
                self.params['translations']['REL_MZ_RANGE'],
            'REL_I_RANGE':
                self.params['translations']['rel_intensity_range'],
            'INTERNAL_PRECISION':
                self.params['translations']['internal_precision'],
            'MAX_MOLECULES_PER_MATCH_BIN':
                self.params['translations']['max_molecules_per_match_bin'],
            'SILAC_AAS_LOCKED_IN_EXPERIMENT':
                self.params['translations']['silac_aas_locked_in_experiment'],
            'BUILD_RESULT_INDEX':
                self.params['translations']['build_pyqms_result_index'],
            'MACHINE_OFFSET_IN_PPM':
                self.params['translations']['machine_offset_in_ppm'],
            'FIXED_LABEL_ISOTOPE_ENRICHMENT_LEVELS':
                self.params['translations'][
                    'fixed_label_isotope_enrichment_levels'],
            # 'COLORS'                                  :
            #     self.params['pyQms_colors'],
        }

        fixed_labels = {}
        for fixed_mod in self.params['mods']['fix']:
            aa = fixed_mod['aa']
            comp = fixed_mod['composition']
            name = fixed_mod['name']
            if not fixed_labels.get(aa, False):
                fixed_labels[aa] = []
            fixed_labels[aa].append(
                {
                    'element_composition': comp,
                    'evidence_mod_name': name
                }
            )

        out = main(
            mzml_file=mzml_files,
            output_file=output_file,
            fixed_labels=fixed_labels,
            evidence_files=self.params['translations'][
                'quantification_evidences'],
            molecules=self.params['translations']['molecules_to_quantify'],
            rt_border_tolerance=self.params[
                'translations']['rt_border_tolerance'],
            label=self.params['translations']['label'],
            label_percentile=self.params['translations']['label_percentile'],
            min_charge=self.params['translations']['precursor_min_charge'],
            max_charge=self.params['translations']['precursor_max_charge'],
            evidence_score_field=self.params[
                'translations']['evidence_score_field'],
            ms_level=self.params['translations']['ms_level'],
            trivial_names=self.params['translations']['pyqms_trivial_names'],
            pyqms_params=pyqms_params,
            verbose=self.params['translations']['pyqms_verbosity'],
            pickle_name=pickle_file
        )

        self.print_execution_time(tag='execution')
        return out



#!/usr/bin/python

# ./combined_FDR2.py /media/lukas/storage/uni/uSearch/combined_FDR --label 15N --cutoff 0.01

from __future__ import print_function
from __future__ import division

from collections import OrderedDict
from collections import defaultdict as ddict
from itertools import compress, product
from csv import DictReader, DictWriter
from pprint import pprint
import argparse
import glob
import os.path



def geometric_mean ( *values ):
    '''
    (without fancy imports...)
    n = number of values
    geo. mean = nth root of all values multiplied with each other
    '''
    product = 1
    for val in values:
        product *= val
    n = len(values)
    
    return product ** (1 / n)  # nth root of product


def calc_intercept_and_slope (a, b):
    '''
    Given two points in 2D-space (tuples a and b), calculates
    the slope (m) and intercept (b) of the line connecting
    both points. Required for FDR Score calculcation.
    Returns intercept and slope in a tuple.
    '''
    # calculating the slope (m) between two points
    if b[0] - a[0] == 0:
        slope = 0
    else:
        slope = \
        (
            ( b[1] - a[1] )
            /
            ( b[0] - a[0] )
        )
    # calculating the intercept (b) of that slope (b = -m * x + y):
    intercept = -slope*a[0] + a[1]
    
    # prevent PEPs above 1 because that would not make sense.
    #if intercept > 1:
        #intercept = 1

    return (intercept, slope)


def all_combinations( iterable ):
    '''
    example:
    
    input:  [ 'xtandem', 'msgf', 'omssa' ]
    output: [{'omssa'}, {'msgf'}, {'msgf', 'omssa'}, {'xtandem'}, {'xtandem', 'omssa'}, {'msgf', 'xtandem'}, {'msgf', 'xtandem', 'omssa'}]
    from http://stackoverflow.com/questions/464864/python-code-to-pick-out-all-possible-combinations-from-a-list
    '''
    out = [ set(compress(iterable,mask)) for mask in product(*[[0,1]]*len(iterable)) ]
    return out[1:]


def calc_FDR ( PSM_count, false_positives ):
    '''
    FDR Method 2 (Käll et al. 2008) as explained by Jones et al. 2009
    '''
    true_positives  = PSM_count - (2 * false_positives)
    if true_positives <= 0:  # prevent FDR above 1. Not sure if this is done?
        return 1.0
    FDR = false_positives / ( false_positives + true_positives)
    return FDR


def calc_FDR_score( e_value, gradient, intercept ):
    '''
    calculating the FDR Score according to Jones et al. 2009
    Algorithm 1.4.b.
    '''
    
    FDR_score = ( e_value * gradient ) + intercept
    
    return FDR_score



class MultiScorer(object):
    # attributes that are shared across all input CSV instances:
    csv_instances  = []  # list of existing class instances (one for each engine)
    AFS_lookup     = {}  # will store the average fdr score of each PSM
    summary_output = []  # will be populated with row_dicts
    known_engines = [
        "tandem", "omssa", "msgf", "msamanda", "myrimatch",
    ]
    unknown_engine_count = 0
    # if an engine name can't be identified from the input csv, it will be called "unknown_engine_{count}"

    def __init__( self, csv_path, info=None ):
        MultiScorer.csv_instances.append( self )  # add this class instance to the list of instances

        self.csv_path = csv_path

        if info == None:
            self.engine         = self.find_engine_name()
            self.engine_colname = self.engine
        else:
            self.engine         = info["engine_name"]
            self.engine_colname = info["engine_colname"]

        self.rowdict  = self.parse_csv()


    def parse_csv( self ):
        '''
        parses the input csv file and indexes it with PSM tuples
        returns a dict: { ('ASDF', 1234) : row, ... }
        '''
        
        # since we do not need all the column information later,
        # we will not store some information to save a bit of RAM:
        deletable_cols = ["PSM_ID","Pass Threshold"]
        
        fname = os.path.basename( self.csv_path )
        print('[{0:^22}] {1}'.format( self.engine.upper(), fname ))
        
        with open(self.csv_path, 'r') as csv_file:

            # skip csv comments / preamble:
            for line in csv_file:
                if not line.startswith('#'):
                    header = line.strip().split(',')
                    break

            reader = DictReader(csv_file, fieldnames=header)

            row_dict = {}
            
            for row in reader:

                # save some RAM
                for colname in deletable_cols:
                    if colname in row:
                        del row[ colname ]
                
                # only consider 'the best' PSM for each spectrum
                #if "rank" in row:
                    #if int(row["rank"]) != 1:
                        #continue
                # same effect as 'select-best-PSM.py' from msblender
                
                pep_sequence        = row[ 'Sequence' ]
                modifications       = row[ 'Modifications' ]
                spectrum_identifier = row[ 'Spectrum Title' ]

                # if 'decoy' in row['proteinacc_start_stop_pre_post_;'].lower():
                #     row['Is decoy'] = 'true'

                PSM_tuple        = ( pep_sequence, spectrum_identifier, modifications, row['Is decoy'] )
                row["PSM tuple"] = PSM_tuple
                row["Percolator:q-Value"] = row["q-value"]
                del row["q-value"]
                
                if PSM_tuple not in row_dict:
                    row_dict[ PSM_tuple ] = row
                else:
                    # if there are multiple PSMs with the same sequence and ID
                    # (i.e. different modifications), only the PSM with the lowest
                    # PEP is kept.
                    old_PEP = float( row_dict[ PSM_tuple ]["PEP"] )
                    new_PEP = float( row["PEP"] )
                    if new_PEP < old_PEP:
                        row_dict[ PSM_tuple ] = row

        return row_dict


    def find_engine_name( self ):
        ''' guess the engine name from its filename '''
        fname = os.path.split(self.csv_path)[-1]
        engine_name = False
        for string in MultiScorer.known_engines:
            if string in fname.lower():
                engine_name = string
        if not engine_name:
            MultiScorer.unknown_engine_count += 1
            engine_name = "unknown_engine_" + str(MultiScorer.unknown_engine_count)
        
        return engine_name

    @classmethod
    def _reset( cls ):
        '''
        cleanup after a single run... not sure why this is necessary
        '''
        cls.csv_instances  = []  # list of existing class instances (one for each engine)
        cls.AFS_lookup     = {}  # will store the average fdr score of each PSM
        cls.summary_output = []  # will be populated with row_dicts
        cls.known_engines = [
            "tandem", "omssa", "msgf", "msamanda", "myrimatch",
        ]
        cls.unknown_engine_count = 0


    @classmethod
    def get_AFS( cls, PSM_tuple, engine_scores ):
        '''
        returns the geometric average FDR or PEP.
        attempts to find this value in a lookup dict,
        if not present it will be calculated based on
        the scores of each engine (engine_scores)
        '''
        if PSM_tuple in cls.AFS_lookup:
            return cls.AFS_lookup[ PSM_tuple ]

        else:
            AFS = geometric_mean( *engine_scores )
            cls.AFS_lookup[ PSM_tuple ] = AFS
            return AFS


    @classmethod
    def get_shared_PSM_rowlist( cls, csv_instances, other_csv_instances ):

        PSM_tuple_sets = [ 
            set(eng.rowdict.keys()) for eng in csv_instances
        ]
        other_PSM_tuple_sets = [
            set(eng.rowdict.keys()) for eng in other_csv_instances
        ]

        engines_str = ";".join( sorted( {i.engine_colname for i in csv_instances} ) )
        shared_PSMs = set.intersection( *PSM_tuple_sets ).difference ( *other_PSM_tuple_sets )

        PSM_rowlist = []
        for PSM in shared_PSMs:

            PSM_scores = []
            PSM_decoy_bools = []
            
            potentially_conflicting_colnames = {
                "Retention Time (s)" : set(),
                "Charge"             : set(),
                #"Calc m/z"           : set(),
                #"Exp m/z"            : set(),
                #"Is decoy"           : set(),
            }

            for csv in csv_instances:
                csv_row = csv.rowdict[ PSM ]

                PSM_scores.append( float(csv_row[ 'FDR Score' ]) )  # FINALLY FIXED
                PSM_decoy_bools.append( csv_row[ 'Is decoy' ] )

                # debugging
                for k in potentially_conflicting_colnames.keys():
                    # try:
                    v = csv_row[ k ]
                    # except:
                    #     import pprint
                    #     pprint.pprint(csv_row)
                    #     print(k)
                    #     exit()
                    potentially_conflicting_colnames[k].add( v )
            
            # if one engine thinks a hit is decoy, it probably is...
            PSM_is_decoy = 'true' if 'true' in PSM_decoy_bools else 'false'

            row = {
               'PSM'      : PSM,
               'AFS'      : cls.get_AFS(PSM, PSM_scores),
               
               'Is decoy' : PSM_is_decoy,
               'Engines'  : engines_str,
               
               # debugging information, remove this later (TODO)
               '_individual_scores'      : PSM_scores,
               '_individual_decoy_bools' : PSM_decoy_bools,
            }
                
            for k,v in potentially_conflicting_colnames.items():
                if len(v) > 1:
                    print('[ Warning! ] Conflicting {0} column: {1}'.format( k, str(v) ) )
                # debugging
                row[k] = list(v)[0]
                

            PSM_rowlist.append( row )
    
        return PSM_rowlist


    @classmethod
    def add_estimated_FDR( cls, shared_PSMs ):
        '''
        Adding the estimated FDR (based on # of decoys).
        based on Jones et al. 2009: Algorithm 2, step 3
        careful, input file must be sorted by average score (low->high) already!
        '''
        
        # In certain datasets, decoy hits are not
        # observed at any score threshold for identifications made by all
        # three search engines. To correct for the size of the result set, an
        # artificial decoy hit is added at the end of each data series, such
        # that no identification has a combined FDR/PEP Score = 0.
        artificial_decoy = [ {'Is decoy' : 'true'} ]
        
        PSM_count = 0
        decoy_count = 0

        # Traverse identifications from lowest AFS to highest:
        for row in shared_PSMs + artificial_decoy:
            PSM_count += 1

            if row['Is decoy'] == 'true':
                decoy_count += 1

            # calculate and store the estimated FDR (FDRest) for each identification according to FDR Method 2.
            row['Estimated FDR']  = calc_FDR ( PSM_count, decoy_count )
            #row['_decoy_percent'] = decoy_count / PSM_count


    @classmethod
    def add_qvalues( cls, shared_PSMs ):
        '''
        Adding (engine-combination specific!) q-values (based on estimated FDR).
        based on Jones et al. 2009: Algorithm 2, step 4
        careful, input file must be sorted by average score (high->low) already!
        '''
    
        # 3) Traverse identifications from highest e-value to lowest, storing
        # the lowest estimated FDR (FDR min) that has been observed so far.
        FDRmin = float('inf')  # starts as infinity, ugly but works...
        for row in shared_PSMs:
            # a. For each identification, retrieve FDRest:
            FDRest = row['Estimated FDR']

            # i. If FDRest > FDR min, q-value = FDR min.
            if FDRest > FDRmin:
                q_value = FDRmin
            # ii. Else, q-value = FDRest and FDR min = FDRest.
            else:
                q_value = FDRest
                FDRmin = FDRest

            row['q-Value'] = q_value

    @classmethod
    def add_FDR_score( cls, idents ):
        # 4) Traverse the set of identifications from
        # lowest q-value to highest, identifying a set of step points, where
        # the q-value of the identification changes (q-value[i] > q-value[i-1] ).

        step_point_prev = ( 0.0, 0.0 )  # intercept is (0,0)
        intercept       =   0
        q_val_prev      = 0.0
        gradient        =   0

        for i, row in enumerate(idents):

            # a. At a step-point, calculate the intercept c and gradient g
            # between previous step-point[prev] (e-value[prev], q-value[prev]) and
            # current step-point[curr] (e-value[curr], q-value[curr] ).
            # Since we don't have q-values in out output, we use the percolator PEP
            # instead. It's only used for ordering, so any quality measure should
            # be fine!
            e_val = float( row['PEP'] )
            q_val = float( row['q-Value'] )

            # check if we are currently at a step-point! otherwise keep previous gradient
            # or if we reached the end of the file, this also counts as a step point...
            if q_val != q_val_prev or i == len(idents)-1:
                step_point_curr = ( e_val, q_val )
                
                intercept, gradient = calc_intercept_and_slope( step_point_prev, step_point_curr )

                # we found a new step point and calculated its gradient and intecept,
                # so let's enter the FDR Score for all PSMs between this step point and the previous one!
                # looping backwards and entering FDR Score until we reach the previous step point...
                pos = 0
                while i-pos >= 0:
                    earlier_row = idents[i-pos]
                    if "FDR Score" in earlier_row.keys():
                        # FDR Score is already entered, so we must have reached an 'old' step point
                        break
                    
                    PSM_e_val = float( earlier_row['PEP'] )
                    combined_score = calc_FDR_score ( PSM_e_val, gradient, intercept )
                    earlier_row[ "FDR Score" ] = combined_score
                    
                    pos += 1

                # in the next iteration, this will be the previous step point
                step_point_prev = step_point_curr

            q_val_prev = q_val


    @classmethod
    def add_combined_FDR_score( cls, shared_PSMs ):
        # 4) Traverse the set of identifications from
        # lowest q-value to highest, identifying a set of step points, where
        # the q-value of the identification changes (q-value[i] > q-value[i-1] ).

        qval_sorted_csv = sorted(shared_PSMs,
                key=lambda x: (float(x['q-Value']), x['AFS']) ) # sort csv by q-value (AFS if ties)

        step_point_prev = ( 0.0, 0.0 )  # intercept is (0,0)
        intercept       =   0
        q_val_prev      = 0.0
        gradient        =   0

        for i, row in enumerate(qval_sorted_csv):

            # a. At a step-point, calculate the intercept c and gradient g
            # between previous step-point[prev] (e-value[prev], q-value[prev]) and
            # current step-point[curr] (e-value[curr], q-value[curr] ).
            e_val = float( row['AFS'] )
            q_val = float( row['q-Value'] )

            # check if we are currently at a step-point! otherwise keep previous gradient
            # or if we reached the end of the file, this also counts as a step point...
            if q_val != q_val_prev or i == len(qval_sorted_csv)-1:
                step_point_curr = ( e_val, q_val )
                
                intercept, gradient = calc_intercept_and_slope( step_point_prev, step_point_curr )

                # we found a new step point and calculated its gradient and intecept,
                # so let's enter the FDR/PEP Score for all PSMs between this step point and the previous one!
                # looping backwards and entering FDR/PEP Score until we reach the previous step point...
                pos = 0
                while i-pos >= 0:
                    earlier_row = qval_sorted_csv[i-pos]
                    if 'Combined FDR Score' in earlier_row.keys():
                        # combined PEP score is already entered, so we must have reached an 'old' step point
                        break
                    
                    PSM_e_val = float( earlier_row['AFS'] )
                    combined_score = calc_FDR_score ( PSM_e_val, gradient, intercept )
                    earlier_row[ 'Combined FDR Score' ] = combined_score
                    
                    if earlier_row['Is decoy'] == 'false':
                        cls.current_stats['PSM_count'] += 1
                        if earlier_row['Combined FDR Score'] < cls.cutoff:
                            cls.current_stats['passed_count'] += 1
                        #if earlier_row['AFS'] < cls.cutoff:
                            #cls.current_stats['percolator_passed_count'] += 1
                    
                    cls.summary_output.append( earlier_row )

                    pos += 1

                # in the next iteration, this will be the previous step point
                step_point_prev = step_point_curr

            q_val_prev = q_val


    @classmethod
    def current_engine_combination_status_str( cls, used_engines, other_engines ):
        
        engines_string = ' and '.join( [i.engine for i in used_engines] )
        other_engines_string = ' or '.join( [i.engine for i in other_engines] )
        
        if len(other_engines) == 0:
            return 'Scoring PSMs that were identified by all {0} engines...'.format( len(used_engines) )
        else:
            return 'Scoring PSMs that were identified by {0}, but not by {1}...'.format( engines_string, other_engines_string )


    @classmethod
    def print_PSM_threshold_stats( cls ):
        d = cls.current_stats
        if d['PSM_count'] == 0:  # preventing division by 0 errors
            d['pass_fraction']            = 0
            #d['percolator_pass_fraction'] = 0
        else:
            d['pass_fraction']            = (d['passed_count'] / d['PSM_count'])
            #d['percolator_pass_fraction'] = (d['percolator_passed_count'] / d['PSM_count'])
        d['threshold']                    = cls.cutoff
        d['score']                        = 'Combined FDR Score'
        
        s1 = '{pass_fraction:>6.1%} of target PSMs are below {threshold:.1%} {score}.   ({passed_count:>6} /{PSM_count:>7})\n'
        #s2 = '{percolator_pass_fraction:>6.1%} of target PSMs are below {threshold:.1%} mean Percolator PEP.  ({percolator_passed_count:>6} /{PSM_count:>7})\n'
        print( s1.format( **d ) )
        #print( s2.format( **d ) )


    @classmethod
    def write_summary_csv( cls, full_file_path ):

        print( 'Writing summary CSV...' )
        
        all_PSMs = cls.AFS_lookup.keys()
        #header   = ['Raw data location', 'Spectrum ID', 'Charge', 'Sequence', 'proteinacc_start_stop_pre_post_;', 'Is decoy', 'Combined PEP Score', '_individual_decoy_bools', '_individual_scores']
        header_start = [
            'Raw data location', # new
            'Spectrum ID',
            'Spectrum Title', # new
            'Retention Time (s)',
            'Calc m/z', # new
            'Exp m/z', # new
            'Charge',
            'Sequence', 
            'Modifications',
            'Engines',
            'Combined FDR Score', 
            'Average FDR Score',
        ]
        header_end = [
            'Is decoy',
            'Protein ID'
        ]
        header_center = []
        for instance in cls.csv_instances:
            header_center.append( instance.engine_colname + ':PEP' )
            header_center.append( instance.engine_colname + ':FDR' )

        header = header_start + header_center + header_end
        
        output_row_list = []
        
        for raw_row in cls.summary_output:
            # decoys and hits above cutoff are filtered out, if desired
            if cls.filter_cutoff and float(raw_row[ 'Combined FDR Score' ]) > cls.cutoff:
                continue
            if cls.filter_decoys and raw_row['Is decoy'] == 'true':
                continue
            
            PSM = raw_row['PSM']
            row = {}
            
            for key in header:
                if key in raw_row:
                    row[key] = raw_row[key]
            row['Average FDR Score'] = raw_row['AFS']
            #row['q_value']           = raw_row['q-Value']
            #row['fdr_est']           = raw_row['Estimated FDR']

            for instance in cls.csv_instances:
                if PSM in instance.rowdict:
                    row[ instance.engine_colname + ':PEP'] = instance.rowdict[PSM]['PEP']
                    row[ instance.engine_colname + ':FDR'] = instance.rowdict[PSM]['Estimated FDR']
                    for key in header:
                        if key in instance.rowdict[PSM]:
                            row[key] = instance.rowdict[PSM][key]

            output_row_list.append( row )

        # sorting the output csv by quality and then dumping it in a file:
        sorted_output_row_list = sorted(
            output_row_list,
            key = lambda x: ( float(x['Combined FDR Score']), float(x['Average FDR Score']) )
            )
        with open(full_file_path, 'w', newline='') as output_file:
            dict_writer = DictWriter( output_file, header )
            dict_writer.writeheader()
            for output_row in sorted_output_row_list:
                dict_writer.writerow( output_row )

        print( 'Wrote summary CSV to {}\n'.format( full_file_path ) )
        return full_file_path


def main(
    input_file_list,
    directory,
    cutoff          = 0.01,
    file_info       = None,
    output_filename = None,
    filter_decoys   = True,
    filter_cutoff   = True,
    label           = "",
        ):
    
    print( "Parsing CSV files:" )
    
    for csv_path in input_file_list:
        if file_info == None:
            # if no info is provided, engine name and label name will be guessed from the file name
            MultiScorer( csv_path )
        else:
            MultiScorer( csv_path, info = file_info[ csv_path ] )

    MultiScorer.cutoff        = cutoff
    MultiScorer.filter_decoys = filter_decoys
    MultiScorer.filter_cutoff = filter_cutoff
    
    print()

    # algorithm 1 from Jones et al. (2009):
    # for each ident CSV, calculating the FDR Score for each PSM
    for csv_instance in MultiScorer.csv_instances:
        ident_csv_unsorted = list( csv_instance.rowdict.values() )
        ident_csv = sorted(
            ident_csv_unsorted,
            key=lambda x: ( float(x['PEP']), float(x['Percolator:q-Value']) )
        )

        MultiScorer.add_estimated_FDR( ident_csv )
        MultiScorer.add_qvalues( reversed(ident_csv) )
        MultiScorer.add_FDR_score( ident_csv )

        # we calculated FDR Scores, so let's add them to the CSV's rowdict!
        for row in ident_csv:
            csv_instance.rowdict[ row['PSM tuple'] ]['FDR Score'] = \
                row['FDR Score']
                
    
    # algorithm 2 from Jones et al. (2009):
    # re-calculating the FDR-scores for each set of identifications that
    # was made by a particular combination of search engines:
    for engine_combo in all_combinations( MultiScorer.csv_instances ):
        other_engines = [i for i in MultiScorer.csv_instances if i not in engine_combo]
        
        MultiScorer.current_stats  = {     # reset stats counter for the current engine combination...
            'PSM_count'              : 0,
            'passed_count'           : 0,  # how many PSMs passed our combined FDR threshold
            #'percolator_passed_count': 0,  # how many PSMs passed the input percolator PEP threshold
        }  
        
        # printing i.e. 'Scoring PSMs that were identified by omssa & msgf, but not by xtandem...'
        status_string = MultiScorer.current_engine_combination_status_str( engine_combo, other_engines )
        print( status_string, end='\r' )
        
        # peptide identifications made by a particular combination of search engines:
        shared_PSMs_unordered = MultiScorer.get_shared_PSM_rowlist( engine_combo, other_engines )
        
        # Order identifications according to AFS (= geometric mean fdr score) within that set:
        shared_PSMs = sorted(shared_PSMs_unordered, key=lambda x: float(x['AFS']) )
        
        MultiScorer.add_estimated_FDR( shared_PSMs )
        MultiScorer.add_qvalues( reversed(shared_PSMs) )
        MultiScorer.add_combined_FDR_score( shared_PSMs )
        
        print( status_string + ' done!' )
        
        MultiScorer.print_PSM_threshold_stats()


    if output_filename == None:
        output_filename = 'combined_FDR_summary_{0}.csv'.format( label )
    else:
        # make sure we don't add the directory twice
        output_filename = os.path.basename( output_filename )

    full_file_path = os.path.join(directory, output_filename )
    summary_csv = MultiScorer.write_summary_csv( full_file_path )
    
    MultiScorer._reset()  # cleaning up previous csv instances
    
    return summary_csv


if __name__ == '__main__':
    
    # parsing command line arguments:
    parser = argparse.ArgumentParser()
    parser.add_argument('directory',
        help='the directory in which to look for \'*complete_v.csv\'-files')
    parser.add_argument('-C', '--cutoff', default=0.01,
        help='the desired combined FDR cutoff, as a fraction', type=float)
    parser.add_argument('-L', '--label',
        help='a substring that allows label identification from the csv file name')
    args = parser.parse_args()

    input_file_list = []
    found_csv = False
    for csv_path in glob.glob(os.path.join(args.directory, '*complete_v.csv')):
        if args.label in os.path.split(csv_path)[-1]:
            input_file_list.append( csv_path )
            found_csv = True
    if not found_csv:
        print( 'Error: No CSV file that ends in \'*complete_v.csv\' and contains label identifier \'{0}\' was found in directory \'{1}\'.'.format(args.label, args.directory) )



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
import os
import pprint
import ursgal.resources.platform_independent.arc_independent.combine_pep_1_0_0.combine_pep_1_0_0 as cpep


TESTS = [
    {
        'iterable': [1,2,3,4,5],
        'window_len': 3,
        'out_windows': [
            (1, [1, 2]),
            (2, [1, 2, 3]),
            (3, [2, 3, 4]),
            (4, [3, 4, 5]),
            (5, [4, 5]),
        ],
    },

    {

        'iterable': [1,0,0,1,1],
        'window_len': 3,
        'out_windows': [
            (1, [1, 0]),
            (0, [1, 0, 0]),
            (0, [0, 0, 1]),
            (1, [0, 1, 1]),
            (1, [1, 1]),
        ],
    },

    {
        'iterable': [1,2,3,4,5],
        'window_len': 2,
        'out_windows': [
            (1, [1, 2]),
            (2, [1, 2, 3]),
            (3, [2, 3, 4]),
            (4, [3, 4, 5]),
            (5, [4, 5]),
        ],
    },

    {
        'iterable': [1,2,3,4,5],
        'window_len': 5,
        'out_windows': [
            (1, [1, 2, 3]),
            (2, [1, 2, 3, 4]),
            (3, [1, 2, 3, 4, 5]),
            (4, [2, 3, 4, 5]),
            (5, [3, 4, 5]),
        ],
    },

    {
        'iterable': [1,2,3,4,5],
        'window_len': 7,
        'out_windows': [
            (1, [1, 2, 3, 4]),
            (2, [1, 2, 3, 4, 5]),
            (3, [1, 2, 3, 4, 5]),
            (4, [1, 2, 3, 4, 5]),
            (5, [2, 3, 4, 5]),
        ],
    },

    {
        'iterable': [1,2,3,4,5],
        'window_len': 99999,
        'out_windows': [
            (1, [1, 2, 3, 4, 5]),
            (2, [1, 2, 3, 4, 5]),
            (3, [1, 2, 3, 4, 5]),
            (4, [1, 2, 3, 4, 5]),
            (5, [1, 2, 3, 4, 5]),
        ],
    },

    {
        'iterable': [1,2,3,4,5],
        'window_len': 1,
        'out_windows': [
            (1, [1]),
            (2, [2]),
            (3, [3]),
            (4, [4]),
            (5, [5]),
        ],
    },

    {
        'iterable': [1,2,3,4,5,6,7],
        'window_len': 11,
        'out_windows': [
            (1, [1, 2, 3, 4, 5, 6]),
            (2, [1, 2, 3, 4, 5, 6, 7]),
            (3, [1, 2, 3, 4, 5, 6, 7]),
            (4, [1, 2, 3, 4, 5, 6, 7]),
            (5, [1, 2, 3, 4, 5, 6, 7]),
            (6, [1, 2, 3, 4, 5, 6, 7]),
            (7, [2, 3, 4, 5, 6, 7]),
        ],
    },

    {
        'iterable': [1],
        'window_len': 999,
        'out_windows': [
            (1, [1]),
        ],
    },

]


def check_sliding_window_test():
    for test_id, test_dict in enumerate(TESTS):
        yield check_sliding_window, test_dict


def check_sliding_window( test_dict ):

    print('\nInput list:', str(test_dict['iterable']), '; win_size:', test_dict['window_len'])

    window_slider = cpep.sliding_window(
        test_dict['iterable'],
        test_dict['window_len'],
        flexible=False,
    )
    expected_results = test_dict['out_windows']
    i = -1
    for i, win_center_tuple in enumerate(window_slider):

        windowsum = win_center_tuple[0]

        center = win_center_tuple[1]
        expected_window = expected_results[i][1]
        expected_center = expected_results[i][0]


        error_str = '\n  Expected:\n\t{0}\t{1}\t{2}\n  Got:\n\t{3}\t{4}'.format(
            expected_center, sum(expected_window), expected_window,
            center, windowsum
        )
        assert center == expected_center, error_str
        assert windowsum == sum(expected_window), error_str
        print(center, windowsum)

    assert i+1 == len(expected_results), '''\n
    Sliding window yielded {0} iterations,
    but {1} iterations were expected.\n
    Test case:\n{2}
    '''.format(i+1, len(expected_results),
               pprint.pformat(test_dict))

if __name__ == '__main__':
    for test_id, test_dict in enumerate(TESTS):
        check_sliding_window( test_dict )



#!/usr/bin/env python3
# encoding: utf-8

import ursgal

TESTS = [
    {
        'input'  : [
            {'XTandem': '34', 'Spectrum ID': '5', 'PEP': '3.5'},
            {'XTandem': '1', 'Spectrum ID': '5', 'PEP': '2'}
        ],
        'output' : {
            'XTandem': 34, 'Spectrum ID': '5', 'PEP': '3.5;2'
        },
        'type_merge' : 'max_value'
    },
    {
        'input'  : [
            {'XTandem': '34', 'Spectrum ID': '5', 'PEP': '3.5'},
            {'XTandem': '1', 'Spectrum ID': '5', 'PEP': '2'}],
        'output' : {
            'XTandem': 1, 'Spectrum ID': '5', 'PEP': '3.5;2'
        },
        'type_merge' : 'min_value'
    },
    {
        'input'  : [
            {'XTandem': '34', 'Spectrum ID': '5', 'PEP': '3.5'},
            {'XTandem': '25', 'Spectrum ID': '5', 'PEP': '2'}
        ],
        'output' : {
            'XTandem': 29.5, 'Spectrum ID': '5', 'PEP': '3.5;2'
        },
        'type_merge' : 'avg_value'
    },
    {
        'input'  : [
            {'XTandem': '34', 'Spectrum ID': '5', 'PEP': '3.5'},
            {'XTandem': '1', 'Spectrum ID': '5', 'PEP': '2'},
            {'XTandem': '1', 'Spectrum ID': '5', 'PEP': '2'}
        ],
        'output' : {
            'XTandem': '1', 'Spectrum ID': '5', 'PEP': '3.5;2;2'
        },
        'type_merge' : 'most_frequent'
    },
    {
        'input'  : [
            {'XTandem': '34', 'Spectrum ID': '5', 'PEP': '3.5'},
            {'XTandem': '34', 'Spectrum ID': '5', 'PEP': '2'},
            {'XTandem': '1', 'Spectrum ID': '5', 'PEP': '2'},
            {'XTandem': '1', 'Spectrum ID': '5', 'PEP': '2'},
            {'XTandem': '2', 'Spectrum ID': '5', 'PEP': '2'}

        ],
        'output' : {
            'XTandem': '34;1', 'Spectrum ID': '5', 'PEP': '3.5;2;2;2;2'
        },
        'type_merge' : 'most_frequent'
    },

]


def merge_rowdicts_test():
    for test_id, test_dict in enumerate(TESTS):
        yield merge_rowdicts, test_dict


def merge_rowdicts(test_dict):
    output = ursgal.ucore.merge_rowdicts(
        list_of_rowdicts=test_dict['input'],
        joinchar=';',
        psm_colnames_to_merge_multiple_values={
            'XTandem': test_dict['type_merge']
        },
    )

    print(output, test_dict)
    assert output == test_dict['output'], '''
        merge_rowdicts failed!
        expected output:
        {0}
        produced putput:
        {1}'''.format(
        test_dict['output'],
        output
    )


if __name__ == '__main__':
    for test_id, test_dict in enumerate(TESTS):
        merge_rowdicts(test_dict)



#!/usr/bin/env python
import ursgal
import os
import subprocess

class uninovo_alpha_052013( ursgal.UNode ):
    """
    UniNovo UNode
    http://proteomics.ucsd.edu/Software/UniNovo/

    Reference:
    Jeong K, Kim S, Pevzner PA (2013): UniNovo: a universal tool for de novo peptide sequencing.
    """
    META_INFO = {
        # see http://proteomics.ucsd.edu/Software/UniNovo/#Downloads
        'edit_version'      : 1.00,
        'name'              : 'UniNovo',
        'version'           : 'alpha.052013',
        'release_date'      : '2013-5-20',
        'engine_type' : {
            'de_novo_search_engine' : True,
        },
        'input_extensions'  : ['.mgf'],
        'output_extensions' : ['.den'],
        'create_own_folder' : True,
        'in_development'    : True,
        'include_in_git'    : None,
        'distributable'      : False,
        'utranslation_style' : 'uninovo_style_1',
        'engine' : {
            'platform_independent' : {
                'arc_independent' : {
                    'exe' : 'UniNovo.jar',
                    'url' : 'http://proteomics.ucsd.edu/Software/UniNovo/UniNovo.20130520.zip',
                    # 'zip_md5'        : '',
                    # 'additional_exe' : [],
                },
            },
        },
        'citation' : \
            'Jeong K, Kim S, Pevzner PA (2013): UniNovo: a universal tool for '\
            'de novo peptide sequencing.',
    }

    def __init__(self, *args, **kwargs):
        super(uninovo_alpha_052013, self).__init__(*args, **kwargs)
        pass

    def preflight( self ):
        '''
        Formatting the command line via self.params

        Returns:
                dict: self.params
        '''

        self.params['mgf_input_file'] = os.path.join(
            self.params['input_dir_path'],
            self.params['file_root'] + '.mgf'
        )

        self.params['output_file_incl_path'] = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )

        print(
            '''
            [ WARNING ] precursor_mass_tolerance_plus and precursor_mass_tolerance_minus
            [ WARNING ] need to be combined for pyQms (use of symmetric tolerance window).
            [ WARNING ] The arithmetic mean is used.
            '''
        )
        self.params['precursor_mass_tolerance'] = ( float(self.params['precursor_mass_tolerance_plus']) + \
                                                    float(self.params['precursor_mass_tolerance_minus']) ) \
                                                / 2.0

        self.params[ 'command_list' ] = [
            'pushd', self.exe.rstrip('UniNovo.jar'), '&&', #need to specify this because UniNovo searches the Pars directory in the shell working directory

            'java', '-Xmx{java_-Xmx}'.format( **self.params), '-jar', self.exe, # path 2 executable
            '-i', '{mgf_input_file}'.format( **self.params), # SpectrumFile (*.mzXML, *.mgf, *.ms2)
            '-o', '{output_file_incl_path}'.format(**self.params).rstrip('.den'), # output file prefix - UniNovo will output file named as "prefix.den"
            '-pt', '{precursor_mass_tolerance}{precursor_mass_tolerance_unit}'.format(**self.params), # precursor ion tolerance (ending in ppm or Da)
            '-t', '{frag_mass_tolerance}{frag_mass_tolerance_unit}'.format(**self.params), # ion tolerances (ending in ppm or Da)
            '-f', '{frag_method}'.format(**self.params), # fragmentation method (CID/ETD/HCD) - if -par option is specified, this option will be ignored
            '-e', '{enzyme}'.format(**self.params), # enzyme applied (0: No enzyme specificity, 1: Trypsin (default), 2: LysC)
            '-c', '{uninovo_num_13C}'.format(**self.params), # number of 13C considered (default : 0)
            #'-l', '{min_pep_length}'.format(**self.params), # minimum length of reconstructions (default : 5), in small datasets UniNovo might produce empty outputfiles, so the whole script crashes
            '-n', '{num_match_spec}'.format(**self.params), # number of de novo sequences per one spectrum (1-200 : default 100)
            '-acc', '{uninovo_accuracy}'.format(**self.params),# set accuracy threshold (0.0-0.9 : default 0.8)
            '-g', '{uninovo_num_mass_gaps}'.format(**self.params), # number of possible mass gaps per each sequence (2-10 : default 10)
            #'-par', {trained_par_file}.format(**self.params), # use user trained parameter file (see below to see how to train UniNovo)

            '&&','popd', #going back to the previous shell working directory
        ]
        #print(self.params[ 'command_list' ])
        return self.params

    def _execute(self):
        '''
        The _execute unode function

        Executes the unode executable via shell.

        Note: internal function
            Unodes that do not require execution via shell
            redefine the _execute() function in their engine
            class.

        Returns:
            None
        '''
        self.print_info('Executing command list ...', caller='eXecution')
        assert 'command_list' in self.params.keys(), '''
  No command_list was found in self.params. Convention is to define
  the command list during preflight in the uNode Engine class code or,
  alternatively, redefine uNode._execute() in the engine class altogether
        '''
        self.time_point(tag = 'execution')
        execute_answer = []
        proc = None

        if len(self.params['command_list']) != 0:
            proc = subprocess.Popen(
                self.params['command_list'],
                stdout = subprocess.PIPE,
                shell = True
            )
        else:
            print('Command list is empty, nothing to do here...')
            print('_execute failed ....', self.params['command_list'])
            execute_answer.append( 'Command list is empty' )
            self.execute_return_code = 500

        if proc is not None:
            for line in proc.stdout:
                line_decoded = line.strip().decode('utf')
                print( line_decoded )
                execute_answer.append( line_decoded )

            # catching the executable's exit code to detect crashes:
            proc.communicate()[0]
            self.execute_return_code = proc.returncode
            assert self.execute_return_code in [0, None], '''
  \n{0} crashed!

  The executable
    {1}
  terminated with Error code {2} .
  Inspect the printouts above for possible causes and verify that all input files are valid.
            '''.format( self.engine, os.path.relpath(self.exe), self.execute_return_code)

        self.print_execution_time(tag='execution')
        return

    def postflight( self ):
        return




#!/usr/bin/env python3
# encoding: utf-8
import ursgal
import sys
import glob
import os


def main(folder=None, profile=None, target_decoy_database=None):
    '''
    An example test script to search all mzML files which are present in the
    specified folder. The search is currently performed on 4 search engines
    and 2 validation engines.

    The machine profile has to be specified as well as the target-decoy
    database.

    usage:

        ./do_it_all_folder_wide.py <mzML_folder> <profile> <target_decoy_database>


    Current profiles:

        * 'QExactive+'
        * 'LTQ XL low res'
        * 'LTQ XL high res'

    '''
    # define folder with mzML_files as sys.argv[1]
    mzML_files = []
    for mzml in glob.glob(os.path.join('{0}'.format(folder), '*.mzML')):
        mzML_files.append(mzml)

    mass_spectrometer = profile

    # We specify all search engines and validation engines that we want to use in a list
    # (version numbers might differ on windows or mac):
    search_engines = [
        'omssa',
        'xtandem_vengeance',
        'msgfplus_v2016_09_16',
        # 'msamanda_1_0_0_6300',
        # 'myrimatch_2_1_138',
    ]

    validation_engines = [
        'percolator_2_08',
        'qvality',
    ]

    # Modifications that should be included in the search
    all_mods = [
        'C,fix,any,Carbamidomethyl',
        'M,opt,any,Oxidation',
        # 'N,opt,any,Deamidated',
        # 'Q,opt,any,Deamidated',
        # 'E,opt,any,Methyl',
        # 'K,opt,any,Methyl',
        # 'R,opt,any,Methyl',
        '*,opt,Prot-N-term,Acetyl',
        # 'S,opt,any,Phospho',
        # 'T,opt,any,Phospho',
        # 'N,opt,any,HexNAc'
    ]

    # Initializing the Ursgal UController class with
    # our specified modifications and mass spectrometer
    params = {
        'database': target_decoy_database,
        'modifications': all_mods,
        'csv_filter_rules': [
            ['Is decoy', 'equals', 'false'],
            ['PEP', 'lte', 0.01],
        ]
    }

    uc = ursgal.UController(
        profile=mass_spectrometer,
        params=params
    )

    # complete workflow:
    # every spectrum file is searched with every search engine,
    # results are validated (for each engine seperately),
    # validated results are merged and filtered for targets and PEP <= 0.01.
    # In the end, all filtered results from all spectrum files are merged
    for validation_engine in validation_engines:
        result_files = []
        for spec_file in mzML_files:
            validated_results = []
            for search_engine in search_engines:
                unified_search_results = uc.search(
                    input_file=spec_file,
                    engine=search_engine,
                )
                validated_csv = uc.validate(
                    input_file=unified_search_results,
                    engine=validation_engine,
                )
                validated_results.append(validated_csv)

            validated_results_from_all_engines = uc.execute_misc_engine(
                input_file=validated_results,
                engine='merge_csvs_1_0_0',
            )
            filtered_validated_results = uc.execute_misc_engine(
                input_file=validated_results_from_all_engines,
                engine='filter_csv_1_0_0',
            )
            result_files.append(filtered_validated_results)

        results_all_files = uc.execute_misc_engine(
            input_file=result_files,
            engine='merge_csvs_1_0_0',
        )

if __name__ == '__main__':
    if len(sys.argv) < 3:
        print(main.__doc__)
        sys.exit(1)
    main(
        folder=sys.argv[1],
        profile=sys.argv[2],
        target_decoy_database=sys.argv[3],
    )



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
import glob
import os.path
import sys
import time
import copy


def main():
    '''

    Example script to compare UPeptideMapper v3 vs v4 results.


    usage:
        ./validate_upeptide_mapper_v3_vs_v4.py

    '''

    input_params = {
        'database': os.path.join(
            os.pardir,
            'example_data',
            'Creinhardtii_281_v5_5_CP_MT_with_contaminants_target_decoy.fasta'
        ),
        'http_url': 'https://www.sas.upenn.edu/~sschulze/Creinhardtii_281_v5_5_CP_MT_with_contaminants_target_decoy.fasta',
        'http_output_folder': os.path.join(
            os.pardir,
            'example_data',
        )
    }

    uc = ursgal.UController(
        params=input_params
    )

    if os.path.exists(input_params['database']) is False:
        uc.fetch_file(
            engine='get_http_files_1_0_0'
        )
    print('Parsing fasta and digesting sequences')
    peptides = set()
    max_number_peptides = 1000000000
    digest_start = time.time()
    for fastaID, sequence in ursgal.ucore.parse_fasta(open(input_params['database'], 'r')):
        tryptic_peptides = ursgal.ucore.digest(
            sequence,
            ('KR', 'C'),
            no_missed_cleavages=True
        )
        for p in tryptic_peptides:
            if 6 <= len(p) <= 40:
                if len(peptides) > max_number_peptides:
                    break
                peptides.add(p)
    print(
        'Parsing fasta and digesting sequences took {0:1.2f} seconds'.format(
            time.time() - digest_start
        )
    )

    # print(peptides)
    upapa_class = uc.unodes['upeptide_mapper_1_0_0']['class'].import_engine_as_python_function(
        'UPeptideMapper_v3'
    )
    print('Buffering fasta and mapping {0} peptides with v3'.format(
        len(peptides)))
    peptide_mapper = upapa_class(input_params['database'])
    fasta_lookup_name = peptide_mapper.fasta_name
    args = [
        list(peptides),
        fasta_lookup_name
    ]
    start_time = time.time()
    v3_p2p_mappings = peptide_mapper.map_peptides(*args)
    print('UPeptideMapper v3 mapper took {0}s'.format(
        time.time() - start_time))
    print('Done')
    v3_p2p_mappings = copy.deepcopy(v3_p2p_mappings)

    upapa_class = uc.unodes['upeptide_mapper_1_0_0']['class'].import_engine_as_python_function(
        'UPeptideMapper_v4'
    )
    print('Buffering fasta and mapping {0} peptides with v4'.format(
        len(peptides)))
    peptide_mapper = upapa_class(input_params['database'])
    args = [
        list(peptides),
    ]
    start_time = time.time()
    v4_p2p_mappings = peptide_mapper.map_peptides(*args)
    print('UPeptideMapper v4 mapper took {0}s'.format(
        time.time() - start_time))
    print('Done')

    assert len(v3_p2p_mappings.keys()) == len(v4_p2p_mappings.keys())
    assert list(sorted(v3_p2p_mappings.keys())) == list(
        sorted(v4_p2p_mappings.keys()))
    compare_keys = [
        'start',
        'end',
        'pre',
        'post',
        'id',
    ]
    num_peps = len(v3_p2p_mappings.keys())
    for ppos, peptide in enumerate(list(sorted(v3_p2p_mappings.keys()))):
        v3_maps = sorted([(d['id'], d['start'], d)
                          for d in v3_p2p_mappings[peptide]])
        v4_maps = sorted([(d['id'], d['start'], d)
                          for d in v4_p2p_mappings[peptide]])
        print(
            'Comparing peptide #{0}/{1}'.format(ppos, num_peps),
            end='\r'
        )
        assert len(v3_maps) == len(v4_maps)

        for pos, (v3_id, v3_start, v3_map_dict) in enumerate(v3_maps):
            v4_id, v4_start, v4_map_dict = v4_maps[pos]
            for key in compare_keys:
                assert v3_map_dict[key] == v4_map_dict[key]


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import ursgal
import os
import tempfile
import time


class _test_node( ursgal.UNode ):
    """_test_node UNode"""
    META_INFO = {
        'edit_version'       : 1.00,
        'name'               : 'test node',
        'version'            : 'alpha',
        'release_date'       : None,
        'engine_type' : {
            '_test'  : True
        },
        'input_extensions'   : ['.txt', '.csv', '.fasta', '.mzML'],
        'output_extensions'  : ['.csv'],
        'output_suffix'      : 'test_node',
        'in_development'     : True,  # do not show in UNode overview
        'include_in_git'     : True,
        'distributable'      : True,
        'utranslation_style' : '_test_node_style_1',
        'engine' : {
            'platform_independent' : {
                'arc_independent' : {
                    'exe' : 'test_node.py',
                },
            },
        },
        'citation' : \
            'TEST/DEBUG: Internal Ursgal UNode for debugging nd testing.',
    }

    def __init__(self, *args, **kwargs):
        super(_test_node, self).__init__(*args, **kwargs)

    def preflight(self):
        self.params['command_list'] = ['sleep', '1']


        # in_path = os.path.join(
        #     self.params['input_dir_path'],
        #     self.params['input_file']
        # )

        out_path = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )
        with open( out_path, 'w') as oof:
            print('Testing 1,2,3 ...', file = oof)
        # # python _test_node.py -i x -o test.txt -t1 a -t2 3
        # self.params['command_list'] = [
        #     'python',
        #     self.exe,
        #     '-i',
        #     in_path,
        #     '-o',
        #     out_path,
        #     '-t1',
        #     '{test_param1}'.format(**self.params['translations']),
        #     '-t2',
        #     '{test_param2}'.format(**self.params['translations']),
        # ]
        time.sleep(2)
        self.print_info(
            'Test UNode: preflight() was executed, slept for 2 seconds!'
        )
        self.print_info(
            'Scheduled command: {command_list}'.format(**self.params )
        )

    def postflight(self):
        time.sleep(1)
        self.print_info("Test UNode: postflight() was executed, slept for 1 second")



#!/usr/bin/env python3
# encoding: utf-8
import ursgal
import pprint
import sys


def main(unode=None):
    '''
    Prints all available parameters for a given UNode, ordered by their tags.

    usage:
        ./show_unode_paramters.py <UNode>

    Example node names:
        * omssa_2_1_9
        * xtandem_vengeance

    '''
    uc = ursgal.UController()
    for unode_key in uc.unodes.keys():
        if unode_key == unode:
            if uc.unodes[ unode_key ].get('available', False):
                print( unode_key)
                pprint.pprint( uc.unodes[unode_key]['class'].UNODE_UPARAMS_GROUPED_BY_TAG)


if __name__ == '__main__':
    main( 
        unode = sys.argv[1]
    )



#!/usr/bin/env python3
# encoding: utf-8
'''

Test the unify_csv function for xtandem engine

'''
import ursgal
import csv
import pickle
import os

psm_defining_colnames = [
    'Spectrum Title',
    'Sequence',
    'Modifications',
    'Mass difference',
    'Charge',
    'Is decoy',
]

input_csv = os.path.join(
    'tests',
    'data',
    'test_merge_duplicate_rows.csv'
)
# output_csv = os.path.join(
#     'tests',
#     'data',
#     'test_merge_duplicate_rows_merged_duplicates.csv'
# )
expected_csv = os.path.join(
    'tests',
    'data',
    'test_merge_duplicate_rows_expected.csv'
)

psm_counter = ursgal.ucore.count_distinct_psms(
    csv_file_path=input_csv,
    psm_defining_colnames=psm_defining_colnames
)

results = ursgal.ucore.merge_duplicate_psm_rows(
    csv_file_path=input_csv,
    psm_counter=psm_counter,
    joinchar=';',
    psm_defining_colnames=psm_defining_colnames,
    overwrite_file=False,
)

merge_line_dicts = []
for line_dict in csv.DictReader(open(results, 'r')):
    merge_line_dicts.append(sorted(line_dict.items()))
sorted_merge_line_dicts = sorted(merge_line_dicts)

expected_line_dicts = []
for line_dict in csv.DictReader(open(expected_csv, 'r')):
    expected_line_dicts.append(sorted(line_dict.items()))
sorted_expected_line_dicts = sorted(expected_line_dicts)

def merge_duplicate_rows_test():
    assert max(psm_counter.values()) > 1, '''
    count_distinct_psms is incorrect:
    No entries > 1
    '''
    for test_id, test_dict_items in enumerate(sorted_merge_line_dicts):
        expected_dict_items = sorted_expected_line_dicts[test_id]
        yield merge_rows, test_dict_items, expected_dict_items


def merge_rows(test_dict_items, expected_dict_items):
    for i, test_item in enumerate(test_dict_items):
        test_key, test_value = test_item
        expected_item = expected_dict_items[i]
        expected_key, expected_value = expected_item
        assert test_key == expected_key, '''
        test_key = {0}
        expected_key = {1}
        '''.format(test_key, expected_key)
        assert test_value == expected_value, '''
        test_value = {0}
        expected_value = {1}
        '''.format(test_value, expected_value)


if __name__ == '__main__':
    print(__doc__)
    for test_id, test_dict in enumerate(sorted(merge_line_dicts)):
        expected_dict = sprted(expcted_line_dicts)[test_id]
        merge_rows(test_dict, expected_dict)



#!/usr/bin/env python
import ursgal
import importlib
import os
import sys
import pickle

class xtandem2csv_1_0_0( ursgal.UNode ):
    """xtandem2csv_1_0_0 UNode"""
    META_INFO = {
        'edit_version'      : 1.00,
        'name'              : 'xtandem2csv',
        'version'           : '1.0.0',
        'release_date'      : None,
        'engine_type' : {
            'converter' : True
        },
        'input_extensions'  : ['.xml', '.xml.gz'],
        'output_extensions' : ['.csv'],
        'output_suffix'     : None,
        'in_development'    : False,
        'include_in_git'    : True,
        'distributable'      : True,
        'utranslation_style' : 'xtandem2csv_style_1',
        'engine' : {
            'platform_independent' : {
                'arc_independent' : {
                    'exe' : 'xtandem2csv_1_0_0.py',
                },
            },
        },
        'citation' : \
            '',
    }

    def __init__(self, *args, **kwargs):
        super(xtandem2csv_1_0_0, self).__init__(*args, **kwargs)

    def _execute( self ):
        '''
        XML result files from X!Tandem are converted to CSV

        Input file has to be a .xml

        Creates a .csv file and returns its path

        '''
        print('[ -ENGINE- ] Executing conversion ..')
        # self.time_point(tag = 'execution')
        xtandem2csv_main = self.import_engine_as_python_function()
        # if self.params['output_file'].lower().endswith('.xml') is False:
        #     raise ValueError('Trying to convert a non-xml file')

        output_file = os.path.join(
                self.params['output_dir_path'],
                self.params['output_file']
            )
        input_file  = os.path.join(
                self.params['input_dir_path'],
                self.params['input_file']
            )

        xtandem2csv_main(
            input_file     = input_file,
            output_file    = output_file,
            decoy_tag      = self.params['translations']['decoy_tag'],
        )

        # self.print_execution_time(tag='execution')
        return output_file



#!/usr/bin/python

import csv
import os
import functools
import operator
import itertools
import collections



def adjust_window_size(desired_window_size, iter_len, minimum=29):
    '''

    Dynamically adjusts the sliding window size depending on the total
    length of values. When there are few values (below 1/5 of the
    window size), the window size is decreased.

    '''
    if desired_window_size < iter_len // 5:
        adjusted_window_size = desired_window_size
    else:
        adjusted_window_size = desired_window_size // 5
        if adjusted_window_size < minimum:
            adjusted_window_size = minimum
    if adjusted_window_size != desired_window_size:

        print('Adjusted window size from {0} to {1} because there '\
              'are only {2} PSMs.'.format(

                  desired_window_size, adjusted_window_size, iter_len
              ))
    return adjusted_window_size



def sliding_window_slow(iterable, window_size, flexible=True):
    '''
    Sliding window generator:
    Slow but readable version using list slicing
    currently not used.
    '''

    if flexible:
        window_size = adjust_window_size(
            window_size, len(iterable)
        )

    if window_size % 2 == 0:
        print('Warning! Window size must be uneven (to determine a '\
              'central value). Adjusted window size from {0} to {1}'\
              '.'.format(window_size, window_size+1))
        window_size += 1


    half_window_size = int((window_size-1)/2)
    for center_i, center_value in enumerate(iterable):
        start_i = center_i - half_window_size
        if start_i < 0:
            start_i = 0
        stop_i = center_i + half_window_size + 1
        yield iterable[start_i:stop_i], center_value


def sliding_window(elements, window_size, flexible=True):
    '''
    Sliding window generator.
    Gives you sliding window functionality without using container
    types (list, deque etc.) to speed it up. Only works for lists of
    numbers. Yields the sum of all numbers in the sliding window
    (= the number of decoys in the sliding window in our case), the
    central number of the sliding window (required for the test only),
    and the current length of the sliding window (= total number of
    PSMs in the sliding window). Used for PEP calculation:
    PEP_of_PSM = (n_decoys_in_window * 2) / n_total_PSMs_in_window
    '''
    if flexible:
        window_size = adjust_window_size(
            window_size, len(elements)
        )

    if window_size % 2 == 0:
        print('Warning! Window size must be uneven (to determine a '\
              'central value). Adjusted window size from {0} to {1}'\
              '.'.format(window_size, window_size+1))
        window_size += 1

    half_window_size = int((window_size-1)/2)

    start_gen, stop_gen = itertools.tee(elements)  # get 2 generators, one that tells
        # us which number to subtract from the back and another one that tells us
        # which number to add at the front of the sliding window

    n_decoys = 0  # keep track of the number of decoys in current sliding window
    current_win_size = 0  # keep track of current window size
    previous_start_i, previous_stop_i = 0, 0  # remember where our sliding window
        # was one iteration earlier (start and stop positions of the sliding window)

    for center_i, center_value in enumerate(elements):
        start_i = center_i - half_window_size
        if start_i < 0:
            start_i = 0
        stop_i = center_i + half_window_size + 1

        if start_i != previous_start_i:
            # I have to substract a number! (the number behind (=left) of the
            # sliding window)
            n_decoys -= next(start_gen)
            current_win_size -= 1
            
        if stop_i != previous_stop_i:
            # I have to add a number! (the next number, the one in front
            # (=right) of the sliding window)
            for i in range(stop_i - previous_stop_i):
                try:
                    n_decoys += next(stop_gen)
                    current_win_size += 1
                except StopIteration:
                    break  # cause StopIteration silently ends for-loops, will be fixed in py3.6 :)

        previous_start_i, previous_stop_i = start_i, stop_i
        yield n_decoys, center_value, current_win_size

class CombinedPEP(object):
    ''''''
    def __init__(self, *args):
        self.psm_dicts = {}
        self.score_dict = {}
        self.input_csv_fieldnames = set()

    @staticmethod
    def row_is_decoy(row):
        '''
        Check if a unified CSV row is a target or decoy PSM.
        Returns True if decoy and False if target.
        '''
        if row['Is decoy'].lower() == 'true':
            is_decoy = True
        elif row['Is decoy'].lower() == 'false':
            is_decoy = False
        else:
            raise Exception('Could not determine whether PSM is decoy or not.')
        if 'decoy_' in row.get('proteinacc_start_stop_pre_post_;', ''):
            is_decoy = True
        return is_decoy

    @staticmethod
    def naive_bayes(prob_list):
        '''
        Combines independent probabilities a, b, c
        like this:

                                a*b*c
        combined_prob = -------------------------
                        a*b*c + (1-a)*(1-b)*(1-c)

        For a straightforward explanation, see
        http://www.paulgraham.com/naivebayes.html
        '''
        # multiplying all probabilities: a*b*c
        multiplied_probs = \
            functools.reduce(operator.mul, prob_list, 1)
        # multiplying the opposite of all probabilities: (1-a)*(1-b)*(1-c)
        multiplied_opposite_probs = \
            functools.reduce(operator.mul, (1-p for p in prob_list), 1)
        return multiplied_probs / (multiplied_probs + multiplied_opposite_probs)

    def add_engine_result_csv(self, input_csv_path, input_engine_name):
        with open(input_csv_path, 'r', encoding='utf8') as csv_obj:
            reader = csv.DictReader(csv_obj)
            for fieldname in reader.fieldnames:
                self.input_csv_fieldnames.add(fieldname)
            self.psm_dicts[input_engine_name] = self.reader_to_psm_dict(
                reader, input_engine_name
            )
        print('Finished parsing {0}'.format(os.path.basename(input_csv_path)))

    def reader_to_psm_dict(self, reader, input_engine_name):
        '''
        Turns a CSV reader object into a dictionary.
        Dict key is a tuple of column values that can be specified by the user
        (usually a combinations of Seq/Mods/Spectrum).
        Dict value is the DictReader row.
        '''
        psm_dict = {}
        for row in reader:
            row_key = self.get_row_key(row, self.columns_for_grouping)
            psm_dict[row_key] = row
        return psm_dict

    @staticmethod
    def list_to_sorted_tuple(l):
        return tuple(sorted(set(l)))

    @staticmethod
    def get_row_key(row, columns_for_grouping):
        return tuple( row[c] for c in columns_for_grouping )

    @staticmethod
    def all_combinations(some_iterable):
        for i in range(len(some_iterable)):
            for combo in itertools.combinations(some_iterable, i+1):
                yield tuple(sorted(combo))

    def generate_psm_to_scores_dict(self, input_engines):
        for engine_combo in self.all_combinations(input_engines):

            engines_not_in_combo = \
                {e for e in input_engines if e not in engine_combo}

            if engines_not_in_combo:
                print('\nScoring PSMs that were found by\n  {0}'\
                    '\nbut not by\n  {1}'.format(
                    ' & '.join(engine_combo),
                    ' or '.join(engines_not_in_combo),
                ))
            else:
                print('\nScoring PSMs that were found by\n  all {0} '\
                    'engines'.format(len(engine_combo)))


            self.score_dict[engine_combo] = {}

            # get all shared PSMs:
            all_PSMs_of_combo_engines = set.intersection(
                *(set(self.psm_dicts[engine].keys()) for engine in engine_combo)
            )


            # remove all PSMs that are found by other engines:
            for other_eng in engines_not_in_combo:
                all_PSMs_of_combo_engines -= set(self.psm_dicts[other_eng].keys())

            if not all_PSMs_of_combo_engines:  # nothing to do here...
                print('There are no PSMs that are unique to this engine combination.')
                continue

            # For every PSM, use naive Bayes to calculate the combined PEP
            # ('Bayes PEP') among all engines and add it to self.score_dict:

            decoy_count_of_intersection = 0
            psm_count_of_intersection = 0
            for PSM_key in all_PSMs_of_combo_engines:
                engine_PEPs = []
                engine_decoy_bools = set()
                for eng in engine_combo:
                    d = self.psm_dicts[eng]
                    engine_PEPs.append(float(d[PSM_key]['PEP']))
                    engine_decoy_bools.add(self.row_is_decoy(d[PSM_key]))

                if len(engine_decoy_bools) != 1:
                    print('Engines don\'t agree whether PSM {0} is decoy or not'\
                          '!'.format(PSM_key))
                    PSM_is_decoy = True
                else:
                    PSM_is_decoy = list(engine_decoy_bools)[0]
                
                psm_count_of_intersection += 1
                if PSM_is_decoy:
                    decoy_count_of_intersection += 1
                    
                bayesPEP = self.naive_bayes(engine_PEPs)

                self.score_dict[engine_combo][PSM_key] = {
                    'Bayes PEP' : bayesPEP,
                    'Is decoy' : PSM_is_decoy,
                }


            print('This set contains {0} PSMs, {1:.1%} of which are decoy.'.format(
                len(all_PSMs_of_combo_engines),
                decoy_count_of_intersection / psm_count_of_intersection,
            ))


            # Use the Bayes PEP to sort all PSMs that are shared by the current
            # combination of engines. Compute the PEP (similar to localized FDR)
            # for each PSM and store it in score_dict:


            psms_sorted_by_bayes_pep = sorted(
                self.score_dict[engine_combo].items(),
                key = lambda kv: kv[1]['Bayes PEP']
            )
            sorted_decoy_bools = [
                kv_tuple[1]['Is decoy'] for kv_tuple in psms_sorted_by_bayes_pep
            ]

            for i, (n_decoys, __, current_win_size) in enumerate(
                sliding_window(sorted_decoy_bools, self.window_size)):

                n_false_positives = 2 * n_decoys
                if n_false_positives > current_win_size:
                    n_false_positives = current_win_size
                intersection_PEP = n_false_positives / current_win_size
                current_PSM_key = psms_sorted_by_bayes_pep[i][0]
                self.score_dict[engine_combo][current_PSM_key]['combined PEP'] = \
                    intersection_PEP

            assert i + 1 == len(sorted_decoy_bools), ('sliding_window() '
                'did not return a sliding window for all PSMs! This '
                'should never happen!')
        return


    def write_output_csv(self, output_csv_path):
        print('\n========== DONE! ===========\nWriting output CSV with combined PEPs '\
            'and Bayes PEPs to\n{} ...\n'.format(os.path.relpath(output_csv_path)))
        new_scores = ['combined PEP', 'Bayes PEP']
        fieldnames = list(self.input_csv_fieldnames) + new_scores + ['engines']
        with open(output_csv_path, 'w', encoding='utf8') as out_obj:
            writer = csv.DictWriter(out_obj, fieldnames=fieldnames)
            writer.writeheader()

            for engine_combo, combo_score_dict in self.score_dict.items():

                for psm_key, score_dict_val in combo_score_dict.items():

                    psm_rows_from_all_engines = []
                    for engine in engine_combo:
                        psm_rows_from_all_engines.append(
                            self.psm_dicts[engine][psm_key]
                        )
                    out_row = self.merge_rowdicts(
                        psm_rows_from_all_engines,
                        sep = self.join_sep,
                    )

                    # add column that lists all engines that found the PSM:
                    out_row['engines'] = self.join_sep.join(engine_combo)
                    # add columns with Bayes PEP and combined PEP:
                    for score_field in new_scores:
                        out_row[score_field] = score_dict_val[score_field]
                    writer.writerow(out_row)
        return

    def merge_rowdicts(self, rows_to_merge, sep=';'):
        '''
        Merges different DictReader rows (=dictionaries) to a single merged
        dictionary. If all values are in agreement, only that value is entered.
        If there are conflicting values, they are joined with a separator
        (i.e. ';').
        '''
        md = {}  # merged rowdict
        for row in rows_to_merge:
            for col_name, field_val in row.items():

                if col_name not in md:
                    # value is not there yet, so add it
                    md[col_name] = field_val
                    continue

                if md[col_name] == field_val:
                    # value is already there, no need to add it
                    continue

                # Different value(s) are already there! join them (i.e. with ';')
                old_vals = set(md[col_name].split(sep))
                old_vals.add(field_val)
                old_vals.discard('')
                md[col_name] = sep.join(sorted(old_vals))
        return md

def main(columns_for_grouping=None, input_csvs=None, output_csv=None,
         input_sep=None, output_sep=None, join_sep=None, pep_colname=None,
         input_engines=None, window_size=None):
    '''
    1. Set parsed attributes from command line as class attribute.
    2. Parse unified input CSV files (with PEPs) and buffer them as
       PSM-to-row dictionaries.
    3. calculate Bayes PEP for each PSM
    4. retrieve all possible engine combinations (like a VennDiagram)
    5. For each engine combination:
        a) retrieve list of shared PSMs
        b) sort PSMs by Bayes PEP
        c) loop over sorted PSMs and calculate combined PEP (sliding window)
    6. Return merged CSV output file with added Bayes PEP and combined PEP
       columns.
    '''
    c = CombinedPEP()

    # Set parsed attributes from command line as class attribute:
    c.columns_for_grouping = c.list_to_sorted_tuple(columns_for_grouping)
    c.input_sep = input_sep
    c.output_sep = output_sep
    c.join_sep = join_sep
    c.pep_colname = pep_colname
    c.window_size = window_size

    # Parse input CSV files and buffer them as PSM-to-row dicts:
    for input_csv, input_engine in zip(input_csvs, input_engines):
        c.add_engine_result_csv(input_csv, input_engine)

    # Calculate Bayes PEP and combined PEP for each PSM:
    c.generate_psm_to_scores_dict(input_engines)

    # Write merged CSV with Bayes PEP and combined PEP column:
    c.write_output_csv(output_csv)
    return

def parse_args(verbose=True):
    '''
    Parses command line arguments and returns them in a dict.
    Only used when executing this script from command line.
    '''
    import argparse
    # no need to import argparse when script is executed by importing
    # main function.

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-i', '--input_csvs', type=str, nargs='+', required=True,
        help='Paths to unified input CSV files (2 or more)')
    parser.add_argument(
        '-c', '--columns_for_grouping', type=str, nargs='+', required=True,
        help='Column names by which the rows should be grouped')
    parser.add_argument(
        '-o', '--output_csv', type=str, help='Output CSV name')
    parser.add_argument(
        '-is', '--input_sep', type=str, default=',',
        help='Input file column delimiter character')
    parser.add_argument(
        '-os', '--output_sep', type=str, default=',',
        help='Output file column delimiter character')
    parser.add_argument(
        '-js', '--join_sep', type=str, default=';',
        help='Delimiter for multiple values in the same field')
    parser.add_argument(
        '-e', '--input_engines', type=str, required=True,
        help='The search engines of each input file (must be same order as input_csvs)')
    parser.add_argument(
        '-w', '--window_size', type=int, default=251,
        help='The size of the sliding window for PEP calculation.')

    args = vars(parser.parse_args())  # convert to dict
    if verbose:
        print('You are using the following settings:')
        for arg, val in sorted(args.items()):
            print('  {0: <13}:  {1}'.format(arg, val))
        print()
    return args


if __name__ == '__main__':
    command_line_args = parse_args()
    main(**command_line_args)




#!/usr/bin/env python
import ursgal
import os
from collections import defaultdict as ddict
import csv
import itertools
import sys

class pglyco_db_2_2_0(ursgal.UNode):
    """
    Unode for pGlyco 2.2.0
    For furhter information see: https://github.com/pFindStudio/pGlyco2

    Note:
        Please download pGlyco 2.2.0 manually from
        https://github.com/pFindStudio/pGlyco2

    Reference:
    Liu MQ, Zeng WF,, Fang P, Cao WQ, Liu C, Yan GQ, Zhang Y, Peng C, Wu JQ,
    Zhang XJ, Tu HJ, Chi H, Sun RX, Cao Y, Dong MQ, Jiang BY, Huang JM, Shen HL,
    Wong CCL, He SM, Yang PY. (2017) pGlyco 2.0 enables precision N-glycoproteomics
    with comprehensive quality control and one-step mass spectrometry
    for intact glycopeptide identification.
    Nat Commun 8(1)
    """

    META_INFO = {
        'edit_version': 1.00,
        'name': 'pGlyco',
        'version': '2.2.0',
        'release_date': '2019-01-01',
        'utranslation_style': 'pglyco_db_style_1',
        'input_extensions': ['.mgf'],
        'output_extensions': ['.csv'],
        'create_own_folder': True,
        'in_development': True,
        'include_in_git': False,
        'distributable': False,
        'engine_type': {
            'protein_database_search_engine': True,
        },
        'engine': {
            'win32' : {
                '64bit' : {
                    'exe'            : 'pGlycoDB.exe',
                    'url'            : '',
                    'zip_md5'        : None,
                    'additional_exe' : [],
                },
            },
        },
        'citation':
        'Liu MQ, Zeng WF,, Fang P, Cao WQ, Liu C, Yan GQ, Zhang Y, Peng C, Wu JQ,'
            'Zhang XJ, Tu HJ, Chi H, Sun RX, Cao Y, Dong MQ, Jiang BY, Huang JM, Shen HL,'
            'Wong CCL, He SM, Yang PY. (2017) pGlyco 2.0 enables precision N-glycoproteomics '
            'with comprehensive quality control and one-step mass spectrometry'
            'for intact glycopeptide identification.'
            'Nat Commun 8(1)'
    }

    def __init__(self, *args, **kwargs):
        super(pglyco_db_2_2_0, self).__init__(*args, **kwargs)
        pass

    def preflight(self):
        '''
        Formatting the command line and writing the param input file via 
        self.params

        Returns:
            dict: self.params
        '''
        self.params['translations']['mgf_input_file'] = os.path.join(
            self.params['input_dir_path'],
            self.params['input_file']
        )
        self.params['translations']['output_file_incl_path'] = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )
        self.param_file_name = os.path.join(
            self.params['translations']['output_file_incl_path'].strip('.csv')
            + '_pGlyco.cfg'
        )
        # self.created_tmp_files.append(self.param_file_name)

        # pprint.pprint(self.params['translations']['_grouped_by_translated_key'])
        # pprint.pprint(self.params)
        # sys.exit(1)
        self.params_to_write = {
            'output_dir_path' : self.params['output_dir_path'],
            'input_file' : self.params['translations']['mgf_input_file'],
        }

        precursor_tolerance = []
        opt_mods = []
        fix_mods = []
        for pglyco_param_name in self.params['translations']['_grouped_by_translated_key'].keys():
            for ursgal_param_name, param_value in self.params['translations']['_grouped_by_translated_key'][pglyco_param_name].items():
                # if pglyco_param_name in write_exclusion_list:
                #     continue
                if pglyco_param_name == 'search_precursor_tolerance':
                    precursor_tolerance.append(param_value)
                elif pglyco_param_name == 'modifications':
                    # fix_total=
                    # fix1=Carbamidomethyl[C]
                    # var_total=
                    # var1=Oxidation[M]
                    # var2=Acetyl[ProteinN-term]
                    sum_opt_mods = 0
                    for mod_dict in self.params['mods']['opt']:
                        '''
                        {'_id': 0,
                          'aa': '*',
                          'composition': {'C': 2, 'H': 2, 'O': 1},
                          'id': '1',
                          'mass': 42.010565,
                          'name': 'Acetyl',
                          'org': '*,opt,Prot-N-term,Acetyl',
                          'pos': 'Prot-N-term',
                          'unimod': True},
                        '''
                        if mod_dict['pos'] == 'Prot-N-term':
                            mod_dict['pos'] = 'ProteinN-term'
                        elif mod_dict['pos'] == 'Prot-C-term':
                            mod_dict['pos'] = 'ProteinC-term'
                        elif mod_dict['pos'] == 'N-term':
                            mod_dict['pos'] = 'AnyN-term'
                        elif mod_dict['pos'] == 'C-term':
                            mod_dict['pos'] = 'AnyC-term'
                        elif mod_dict['pos'] == 'any':
                            pass
                        else:
                            print(
                                '''
                            Unknown positional argument for given modification:
                            {0}
                            pGlyco (or Ursgal) cannot deal with this, please use one of the follwing:
                            any, Prot-N-term, Prot-C-term, N-term, C-term
                            '''.format(mod_dict['org'])
                            )
                            sys.exit(1)
                        sum_opt_mods += 1
                        if 'term' in mod_dict['pos']:
                            if mod_dict['aa'] == '*':
                                mod_dict['aa'] = ''
                            opt_mods.append('var{0}={1}[{2}{3}]'.format(
                                sum_opt_mods,
                                mod_dict['name'],
                                mod_dict['pos'],
                                mod_dict['aa'],
                            ))
                        else:
                            if mod_dict['aa'] == '*':
                                print('''
                                Not sure how to handle this modification in pGlyco:
                                {0}
                                '''.format(mod_dict['org']))
                                sys.exit(1)
                            opt_mods.append('var{0}={1}[{2}]'.format(
                                sum_opt_mods,
                                mod_dict['name'],
                                mod_dict['aa'],
                            ))

                    sum_fix_mods = 0
                    for mod_dict in self.params['mods']['fix']:
                        if mod_dict['pos'] == 'Prot-N-term':
                            mod_dict['pos'] = 'ProteinN-term'
                        elif mod_dict['pos'] == 'Prot-C-term':
                            mod_dict['pos'] = 'ProteinC-term'
                        elif mod_dict['pos'] == 'N-term':
                            mod_dict['pos'] = 'AnyN-term'
                        elif mod_dict['pos'] == 'C-term':
                            mod_dict['pos'] = 'AnyC-term'
                        elif mod_dict['pos'] == 'any':
                            pass
                        else:
                            print(
                                '''
                            Unknown positional argument for given modification:
                            {0}
                            pGlyco (or Ursgal) cannot deal with this, please use one of the follwing:
                            any, Prot-N-term, Prot-C-term, N-term, C-term
                            '''.format(mod_dict['org'])
                            )
                            sys.exit(1)
                        sum_fix_mods += 1
                        if 'term' in mod_dict['pos']:
                            if mod_dict['aa'] == '*':
                                mod_dict['aa'] = ''
                            fix_mods.append('fix{0}={1}[{2}{3}]'.format(
                                sum_fix_mods,
                                mod_dict['name'],
                                mod_dict['pos'],
                                mod_dict['aa'],
                            ))
                        else:
                            if mod_dict['aa'] == '*':
                                print('''
                                Not sure how to handle this modification in pGlyco:
                                {0}
                                '''.format(mod_dict['org']))
                                sys.exit(1)
                            fix_mods.append('fix{0}={1}[{2}]'.format(
                                sum_fix_mods,
                                mod_dict['name'],
                                mod_dict['aa'],
                            ))
                else:
                    self.params_to_write[pglyco_param_name] = param_value
        assert len(precursor_tolerance) == 2, '''
        Both, precursor_mass_tolerance_plus and precursor_mass_tolerance_minus
        need to be given. Received: {0}
        '''.format(precursor_tolerance)
        self.params_to_write['precursor_tolerance'] = sum(precursor_tolerance) / 2.0
        self.params_to_write['sum_fixed_mods'] = sum_fix_mods
        self.params_to_write['sum_opt_mods'] = sum_opt_mods
        self.params_to_write['fix_mods'] = '\n'.join(fix_mods)
        self.params_to_write['opt_mods'] = '\n'.join(opt_mods)
        
        self.write_params_file()

        self.params['command_list'] = [
            self.exe,
            self.param_file_name,
        ]

        return self.params

    def postflight(self):
        # '''
        # Reads pGlyco txt output and write final csv output file.

        # Adds:
        #     * Raw data location, since this can not be added later

        # '''
        pglyco_header = [
            'GlySpec',
            'PepSpec',
            'RawName',
            'Scan',
            'RT',
            'PrecursorMH',
            'PrecursorMZ',
            'Charge',
            'Rank',
            'Peptide',
            'Mod',
            'PeptideMH',
            'Glycan(H,N,A,G,F)',
            'PlausibleStruct',
            'GlyID',
            'GlyFrag',
            'GlyMass',
            'GlySite',
            'TotalScore',
            'PepScore',
            'GlyScore',
            'CoreMatched',
            'CoreFuc',
            'MassDeviation',
            'PPM',
            'GlyIonRatio',
            'PepIonRatio',
            'GlyDecoy',
            'PepDecoy',
        ]

        translated_headers = []
        header_translations = self.UNODE_UPARAMS[
            'header_translations']['uvalue_style_translation']
        for original_header_key in pglyco_header:
            ursgal_header_key = header_translations[original_header_key]
            if ursgal_header_key not in translated_headers:
                translated_headers.append(ursgal_header_key)

        translated_headers += [
            'Raw data location',
        ]

        csv_kwargs = {}
        if sys.platform == 'win32':
            csv_kwargs['lineterminator'] = '\n'
        else:
            csv_kwargs['lineterminator'] = '\r\n'
        csv_writer = csv.DictWriter(
            open(self.params['translations']['output_file_incl_path'], 'w'),
            fieldnames=translated_headers,
            **csv_kwargs
        )

        pglyco_output = os.path.join(
            self.params['output_dir_path'],
            'pGlycoDB-GP.txt'
        )
        translated_headers.insert(0, 'Spectrum Title')
        translated_headers.insert(0, 'Spectrum Title')
        csv_reader = csv.DictReader(
            open(pglyco_output, 'r'),
            fieldnames=translated_headers,
            delimiter='\t',
        )
        # self.created_tmp_files.append(pglyco_output)

        csv_writer.writeheader()
        for n, line_dict in enumerate(csv_reader):
            if n == 0:
                continue
            line_dict['Raw data location'] = os.path.abspath(
                self.params['translations']['mgf_input_file']
            )
            csv_writer.writerow(line_dict)
        return

    def write_params_file(self):
        with open(self.param_file_name, 'w') as io:
            print('''[version]
pGlyco_version=pGlyco_Tag20190101
[flow]
glyco_type=N-Glyco
glycandb=pGlyco.gdb
process={process}
output_dir={output_dir_path}
[protein]
fasta={fasta}
enzyme={enzyme}
digestion=not_support_in_cur_version
max_miss_cleave={max_miss_cleave}
max_peptide_len={max_peptide_len}
min_peptide_len={min_peptide_len}
max_peptide_weight={max_peptide_weight}
min_peptide_weight={min_peptide_weight}
[modification]
fix_total={sum_fixed_mods}
{fix_mods}
max_var_modify_num={max_var_modify_num}
var_total={sum_opt_mods}
{opt_mods}
[search]
search_precursor_tolerance={precursor_tolerance}
search_precursor_tolerance_type={search_precursor_tolerance_type}
search_fragment_tolerance={search_fragment_tolerance}
search_fragment_tolerance_type={search_fragment_tolerance_type}
spec_file_type=mgf
spectrum_total=1
file1={input_file}
'''.format(
                **self.params_to_write),
                file=io
            )
        return



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
import glob
import csv
import os
from collections import defaultdict as ddict
import sys
import re


MQ_OFFSET_TO_FILENAME = [
    (4.71,  'B_D140314_SGSDSsample2_R01_MSG_T0.mzML'),
    (4.98,  'B_D140314_SGSDSsample6_R01_MSG_T0.mzML'),
    (4.9,   'B_D140314_SGSDSsample1_R01_MSG_T0.mzML'),
    (5.41,  'B_D140314_SGSDSsample4_R01_MSG_T0.mzML'),
    (5.78,  'B_D140314_SGSDSsample5_R01_MSG_T0.mzML'),
    (6.01,  'B_D140314_SGSDSsample8_R01_MSG_T0.mzML'),
    (6.22,  'B_D140314_SGSDSsample7_R01_MSG_T0.mzML'),
    (6.83,  'B_D140314_SGSDSsample3_R01_MSG_T0.mzML'),
    (7.61,  'B_D140314_SGSDSsample4_R02_MSG_T0.mzML'),
    (7.59,  'B_D140314_SGSDSsample8_R02_MSG_T0.mzML'),
    (7.93,  'B_D140314_SGSDSsample6_R02_MSG_T0.mzML'),
    (7.91,  'B_D140314_SGSDSsample1_R02_MSG_T0.mzML'),
    (8.23,  'B_D140314_SGSDSsample3_R02_MSG_T0.mzML'),
    (8.33,  'B_D140314_SGSDSsample7_R02_MSG_T0.mzML'),
    (9.2,   'B_D140314_SGSDSsample5_R02_MSG_T0.mzML'),
    (9.4,   'B_D140314_SGSDSsample2_R02_MSG_T0.mzML'),
    (9.79,  'B_D140314_SGSDSsample1_R03_MSG_T0.mzML'),
    (10.01, 'B_D140314_SGSDSsample3_R03_MSG_T0.mzML'),
    (10.03, 'B_D140314_SGSDSsample7_R03_MSG_T0.mzML'),
    (10.58, 'B_D140314_SGSDSsample2_R03_MSG_T0.mzML'),
    (11.1,  'B_D140314_SGSDSsample4_R03_MSG_T0.mzML'),
    (11.21, 'B_D140314_SGSDSsample5_R03_MSG_T0.mzML'),
    (11.45, 'B_D140314_SGSDSsample6_R03_MSG_T0.mzML'),
    (12.19, 'B_D140314_SGSDSsample8_R03_MSG_T0.mzML'),
]

GENERAL_PARAMS = {
    'database': os.path.join(
        os.pardir,
        'example_data',
        'hs_201303_qs_sip_target_decoy.fasta'
    ),
    'modifications': [
        'C,fix,any,Carbamidomethyl',  # Carbamidomethylation
    ],
    'scan_skip_modulo_step': 10,
    'http_url': 'http://www.uni-muenster.de/Biologie.IBBP.AGFufezan/misc/hs_201303_qs_sip_target_decoy.fasta',
    'http_output_folder': os.path.join(
        os.pardir,
        'example_data'
    )
}


def search(input_folder=None):
    '''
    Does the parameter sweep on every tenth MS2 spectrum of the data from
    Bruderer et al. (2015) und X!Tandem Sledgehammer.

    Note:

        Please download the .RAW data for the DDA dataset from peptideatlas.org
        (PASS00589, password: WF6554orn) and convert to mzML.
        Then the script can be executed with the folder with the mzML files as
        the first argument.

    Warning:

        This script (if the sweep ranges are not changed) will perform 10080
        searches which will produce approximately 100 GB output (inclusive mzML
        files)

    usage:

        ./machine_offset_bruderer_sweep.py <folder_with_bruderer_data>

    Sweeps over:

        * machine_offset_in_ppm from -20 to +20 ppm offset
        * precursor mass tolerance from 1 to 5 ppm
        * fragment mass tolerance from -2.5 to 20 ppm

    The search can be very time consuming (depending on your machine/cluster),
    therefor the analyze step can be performed separately by calling analyze()
    instead of search() when one has already performed the searches and wants
    to analyze the results.
    '''

    file_2_target_offset = {}
    for MQ_offset, file_name in MQ_OFFSET_TO_FILENAME:
        file_2_target_offset[file_name] = {'offsets': []}
        for n in range(-20, 20, 2):
            file_2_target_offset[file_name]['offsets'].append(n)

    engine_list = ['xtandem_sledgehammer']

    frag_ion_tolerance_list = [2.5, 5, 10, 20]

    precursor_ion_tolerance_list = [1, 2, 3, 4, 5]

    R = ursgal.UController(
        profile='QExactive+',
        params=GENERAL_PARAMS
    )

    if os.path.exists(R.params['database']) is False:
        R.fetch_file(
            engine='get_http_files_1_0_0'
        )

    prefix_format_string = '_pit_{1}_fit_{2}'
    for mzML_path in glob.glob(os.path.join(input_folder, '*.mzML')):

        mzML_basename = os.path.basename(mzML_path)
        if mzML_basename not in file_2_target_offset.keys():
            continue
        for ppm_offset in file_2_target_offset[mzML_basename]['offsets']:
            R.params['machine_offset_in_ppm'] = ppm_offset
            R.params['prefix'] = 'ppm_offset_{0}'.format(
                int(ppm_offset)
            )
            mgf_file = R.convert(
                input_file=mzML_path,
                engine='mzml2mgf_1_0_0'
            )
            for engine in engine_list:
                for precursor_ion_tolerane in precursor_ion_tolerance_list:
                    for frag_ion_tolerance in frag_ion_tolerance_list:

                        new_prefix = prefix_format_string.format(
                            precursor_ion_tolerane,
                            frag_ion_tolerance
                        )
                        R.params[
                            'precursor_mass_tolerance_minus'] = precursor_ion_tolerane
                        R.params[
                            'precursor_mass_tolerance_plus'] = precursor_ion_tolerane
                        R.params['frag_mass_tolerance'] = frag_ion_tolerance
                        R.params['prefix'] = new_prefix

                        unified_search_result_file = R.search(
                            input_file=mzML_path,
                            engine=engine,
                            force=False,
                        )

    return


def analyze(folder):
    '''

    Parses the result files form search and write a result .csv file which
    contains the data to plot figure 2.

    '''

    R = ursgal.UController(
        profile='QExactive+',
        params=GENERAL_PARAMS
    )
    csv_collector = {}
    ve = 'qvality_2_02'

    sample_regex_pattern = 'sample\d_R0\d'

    sample_2_x_pos_and_mq_offset = {}

    sample_offset_combos = []

    all_tested_offsets = [str(n) for n in range(-20, 21, 2)]

    for pos, (mq_ppm_off, mzML_file) in enumerate(MQ_OFFSET_TO_FILENAME):
        _sample = re.search(sample_regex_pattern, mzML_file).group()
        sample_2_x_pos_and_mq_offset[_sample] = (pos, mq_ppm_off)
        for theo_offset in all_tested_offsets:
            sample_offset_combos.append((_sample, theo_offset))

    for csv_path in glob.glob(os.path.join('{0}'.format(folder), '*', '*_unified.csv')):
        dirname = os.path.dirname(csv_path)
        sample = re.search(sample_regex_pattern, csv_path).group()
        splitted_basename = os.path.basename(csv_path).split('_')
        offset = splitted_basename[2]
        precursor_ion_tolerance = splitted_basename[4]
        frag_ion_tolerance = splitted_basename[6]
        prefix = '_'.join(splitted_basename[:7])

        R.params['machine_offset_in_ppm'] = offset
        R.params['precursor_mass_tolerance_minus'] = precursor_ion_tolerance
        R.params['precursor_mass_tolerance_plus'] = precursor_ion_tolerance
        R.params['frag_mass_tolerance'] = frag_ion_tolerance
        R.params['prefix'] = prefix

        validated_path = csv_path.replace(
            '_unified.csv',
            '_{0}_validated.csv'.format(ve)
        )
        if os.path.exists(validated_path):
            csv_path = validated_path
        else:
            try:
                csv_path = R.validate(
                    input_file=csv_path,
                    engine=ve
                )
            except:
                continue

        pit_fit = (precursor_ion_tolerance, frag_ion_tolerance)

        if pit_fit not in csv_collector.keys():
            csv_collector[pit_fit] = ddict(set)

        csv_key = (sample, offset)

        print('Reading file: {0}'.format(csv_path))
        for line_dict in csv.DictReader(open(csv_path, 'r')):
            if line_dict['Is decoy'] == 'true':
                continue
            if float(line_dict['PEP']) <= 0.01:
                csv_collector[pit_fit][csv_key].add(
                    '{0}{1}'.format(
                        line_dict['Sequence'],
                        line_dict['Modifications']
                    )
                )

    fieldnames = [
        'Sample',
        'pos',
        'MQ_offset',
        'tested_ppm_offset',
        'peptide_count'
    ]

    outfile_name_format_string = 'bruderer_data_ppm_sweep_precursor_mass_tolerance_{0}_fragment_mass_tolerance_{1}.csv'

    for pit_fit in csv_collector.keys():
        with open(outfile_name_format_string.format(*pit_fit), 'w') as io:
            csv_writer = csv.DictWriter(io, fieldnames)
            csv_writer.writeheader()

            # write missing values
            for sample_offset in sample_offset_combos:
                sample, ppm_offset = sample_offset
                if sample_offset not in csv_collector[pit_fit].keys():
                    dict_2_write = {
                        'Sample': sample,
                        'pos': sample_2_x_pos_and_mq_offset[sample][0],
                        'MQ_offset': '',
                        'tested_ppm_offset': ppm_offset,
                        'peptide_count': 0,
                    }
                    csv_writer.writerow(dict_2_write)

            for (sample, ppm_offset), peptide_set in csv_collector[pit_fit].items():
                dict_2_write = {
                    'Sample': sample,
                    'pos': sample_2_x_pos_and_mq_offset[sample][0],
                    'MQ_offset': sample_2_x_pos_and_mq_offset[sample][1] * -1,
                    'tested_ppm_offset': ppm_offset,
                    'peptide_count': len(peptide_set),
                }
                csv_writer.writerow(dict_2_write)
    return

if __name__ == '__main__':
    if len(sys.argv) < 2:
        print(search.__doc__)
        sys.exit(1)
    search(sys.argv[1])
    analyze(sys.argv[1])
